from __future__ import absolute_import, unicode_literals

import logging

from mopidy import listener


logger = logging.getLogger(__name__)


class Backend(object):

    """Backend API

    If the backend has problems during initialization it should raise
    :exc:`mopidy.exceptions.BackendError` with a descriptive error message.
    This will make Mopidy print the error message and exit so that the user can
    fix the issue.

    :param config: the entire Mopidy configuration
    :type config: dict
    :param audio: actor proxy for the audio subsystem
    :type audio: :class:`pykka.ActorProxy` for :class:`mopidy.audio.Audio`
    """

    #: Actor proxy to an instance of :class:`mopidy.audio.Audio`.
    #:
    #: Should be passed to the backend constructor as the kwarg ``audio``,
    #: which will then set this field.
    audio = None

    #: The library provider. An instance of
    #: :class:`~mopidy.backend.LibraryProvider`, or :class:`None` if
    #: the backend doesn't provide a library.
    library = None

    #: The playback provider. An instance of
    #: :class:`~mopidy.backend.PlaybackProvider`, or :class:`None` if
    #: the backend doesn't provide playback.
    playback = None

    #: The playlists provider. An instance of
    #: :class:`~mopidy.backend.PlaylistsProvider`, or class:`None` if
    #: the backend doesn't provide playlists.
    playlists = None

    #: List of URI schemes this backend can handle.
    uri_schemes = []

    # Because the providers is marked as pykka_traversible, we can't get() them
    # from another actor, and need helper methods to check if the providers are
    # set or None.

    def has_library(self):
        return self.library is not None

    def has_library_browse(self):
        return self.has_library() and self.library.root_directory is not None

    def has_playback(self):
        return self.playback is not None

    def has_playlists(self):
        return self.playlists is not None

    def ping(self):
        """Called to check if the actor is still alive."""
        return True


class LibraryProvider(object):

    """
    :param backend: backend the controller is a part of
    :type backend: :class:`mopidy.backend.Backend`
    """

    pykka_traversable = True

    root_directory = None
    """
    :class:`mopidy.models.Ref.directory` instance with a URI and name set
    representing the root of this library's browse tree. URIs must
    use one of the schemes supported by the backend, and name should
    be set to a human friendly value.

    *MUST be set by any class that implements* :meth:`LibraryProvider.browse`.
    """

    def __init__(self, backend):
        self.backend = backend

    def browse(self, uri):
        """
        See :meth:`mopidy.core.LibraryController.browse`.

        If you implement this method, make sure to also set
        :attr:`root_directory`.

        *MAY be implemented by subclass.*
        """
        return []

    def get_distinct(self, field, query=None):
        """
        See :meth:`mopidy.core.LibraryController.get_distinct`.

        *MAY be implemented by subclass.*

        Default implementation will simply return an empty set.

        Note that backends should always return an empty set for unexpected
        field types.
        """
        return set()

    def get_images(self, uris):
        """
        See :meth:`mopidy.core.LibraryController.get_images`.

        *MAY be implemented by subclass.*

        Default implementation will simply return an empty dictionary.
        """
        return {}

    def lookup(self, uri):
        """
        See :meth:`mopidy.core.LibraryController.lookup`.

        *MUST be implemented by subclass.*
        """
        raise NotImplementedError

    def refresh(self, uri=None):
        """
        See :meth:`mopidy.core.LibraryController.refresh`.

        *MAY be implemented by subclass.*
        """
        pass

    def search(self, query=None, uris=None, exact=False):
        """
        See :meth:`mopidy.core.LibraryController.search`.

        *MAY be implemented by subclass.*

        .. versionadded:: 1.0
            The ``exact`` param which replaces the old ``find_exact``.
        """
        pass


class PlaybackProvider(object):

    """
    :param audio: the audio actor
    :type audio: actor proxy to an instance of :class:`mopidy.audio.Audio`
    :param backend: the backend
    :type backend: :class:`mopidy.backend.Backend`
    """

    pykka_traversable = True

    def __init__(self, audio, backend):
        self.audio = audio
        self.backend = backend

    def pause(self):
        """
        Pause playback.

        *MAY be reimplemented by subclass.*

        :rtype: :class:`True` if successful, else :class:`False`
        """
        return self.audio.pause_playback().get()

    def play(self):
        """
        Start playback.

        *MAY be reimplemented by subclass.*

        :rtype: :class:`True` if successful, else :class:`False`
        """
        return self.audio.start_playback().get()

    def prepare_change(self):
        """
        Indicate that an URI change is about to happen.

        *MAY be reimplemented by subclass.*

        It is extremely unlikely it makes sense for any backends to override
        this. For most practical purposes it should be considered an internal
        call between backends and core that backend authors should not touch.
        """
        self.audio.prepare_change().get()

    def translate_uri(self, uri):
        """
        Convert custom URI scheme to real playable URI.

        *MAY be reimplemented by subclass.*

        This is very likely the *only* thing you need to override as a backend
        author. Typically this is where you convert any Mopidy specific URI
        to a real URI and then return it. If you can't convert the URI just
        return :class:`None`.

        :param uri: the URI to translate
        :type uri: string
        :rtype: string or :class:`None` if the URI could not be translated
        """
        return uri

    def change_track(self, track):
        """
        Swith to provided track.

        *MAY be reimplemented by subclass.*

        It is unlikely it makes sense for any backends to override
        this. For most practical purposes it should be considered an internal
        call between backends and core that backend authors should not touch.

        The default implementation will call :meth:`translate_uri` which
        is what you want to implement.

        :param track: the track to play
        :type track: :class:`mopidy.models.Track`
        :rtype: :class:`True` if successful, else :class:`False`
        """
        uri = self.translate_uri(track.uri)
        if uri != track.uri:
            logger.debug(
                'Backend translated URI from %s to %s', track.uri, uri)
        if not uri:
            return False
        self.audio.set_uri(uri).get()
        return True

    def resume(self):
        """
        Resume playback at the same time position playback was paused.

        *MAY be reimplemented by subclass.*

        :rtype: :class:`True` if successful, else :class:`False`
        """
        return self.audio.start_playback().get()

    def seek(self, time_position):
        """
        Seek to a given time position.

        *MAY be reimplemented by subclass.*

        :param time_position: time position in milliseconds
        :type time_position: int
        :rtype: :class:`True` if successful, else :class:`False`
        """
        return self.audio.set_position(time_position).get()

    def stop(self):
        """
        Stop playback.

        *MAY be reimplemented by subclass.*

        Should not be used for tracking if tracks have been played or when we
        are done playing them.

        :rtype: :class:`True` if successful, else :class:`False`
        """
        return self.audio.stop_playback().get()

    def get_time_position(self):
        """
        Get the current time position in milliseconds.

        *MAY be reimplemented by subclass.*

        :rtype: int
        """
        return self.audio.get_position().get()


class PlaylistsProvider(object):

    """
    A playlist provider exposes a collection of playlists, methods to
    create/change/delete playlists in this collection, and lookup of any
    playlist the backend knows about.

    :param backend: backend the controller is a part of
    :type backend: :class:`mopidy.backend.Backend` instance
    """

    pykka_traversable = True

    def __init__(self, backend):
        self.backend = backend

    def as_list(self):
        """
        Get a list of the currently available playlists.

        Returns a list of :class:`~mopidy.models.Ref` objects referring to the
        playlists. In other words, no information about the playlists' content
        is given.

        :rtype: list of :class:`mopidy.models.Ref`

        .. versionadded:: 1.0
        """
        raise NotImplementedError

    def get_items(self, uri):
        """
        Get the items in a playlist specified by ``uri``.

        Returns a list of :class:`~mopidy.models.Ref` objects referring to the
        playlist's items.

        If a playlist with the given ``uri`` doesn't exist, it returns
        :class:`None`.

        :rtype: list of :class:`mopidy.models.Ref`, or :class:`None`

        .. versionadded:: 1.0
        """
        raise NotImplementedError

    def create(self, name):
        """
        Create a new empty playlist with the given name.

        Returns a new playlist with the given name and an URI, or :class:`None`
        on failure.

        *MUST be implemented by subclass.*

        :param name: name of the new playlist
        :type name: string
        :rtype: :class:`mopidy.models.Playlist` or :class:`None`
        """
        raise NotImplementedError

    def delete(self, uri):
        """
        Delete playlist identified by the URI.

        Returns :class:`True` if deleted, :class:`False` otherwise.

        *MUST be implemented by subclass.*

        :param uri: URI of the playlist to delete
        :type uri: string
        :rtype: :class:`bool`

        .. versionchanged:: 2.2
            Return type defined.
        """
        raise NotImplementedError

    def lookup(self, uri):
        """
        Lookup playlist with given URI in both the set of playlists and in any
        other playlist source.

        Returns the playlists or :class:`None` if not found.

        *MUST be implemented by subclass.*

        :param uri: playlist URI
        :type uri: string
        :rtype: :class:`mopidy.models.Playlist` or :class:`None`
        """
        raise NotImplementedError

    def refresh(self):
        """
        Refresh the playlists in :attr:`playlists`.

        *MUST be implemented by subclass.*
        """
        raise NotImplementedError

    def save(self, playlist):
        """
        Save the given playlist.

        The playlist must have an ``uri`` attribute set. To create a new
        playlist with an URI, use :meth:`create`.

        Returns the saved playlist or :class:`None` on failure.

        *MUST be implemented by subclass.*

        :param playlist: the playlist to save
        :type playlist: :class:`mopidy.models.Playlist`
        :rtype: :class:`mopidy.models.Playlist` or :class:`None`
        """
        raise NotImplementedError


class BackendListener(listener.Listener):

    """
    Marker interface for recipients of events sent by the backend actors.

    Any Pykka actor that mixes in this class will receive calls to the methods
    defined here when the corresponding events happen in a backend actor. This
    interface is used both for looking up what actors to notify of the events,
    and for providing default implementations for those listeners that are not
    interested in all events.

    Normally, only the Core actor should mix in this class.
    """

    @staticmethod
    def send(event, **kwargs):
        """Helper to allow calling of backend listener events"""
        listener.send(BackendListener, event, **kwargs)

    def playlists_loaded(self):
        """
        Called when playlists are loaded or refreshed.

        *MAY* be implemented by actor.
        """
        pass



from __future__ import absolute_import, print_function, unicode_literals

import logging
import os
import signal
import sys

import pykka.debug

from mopidy import commands, config as config_lib, ext
from mopidy.internal import encoding, log, path, process, versioning
from mopidy.internal.gi import Gst  # noqa: F401

try:
    # Make GLib's mainloop the event loop for python-dbus
    import dbus.mainloop.glib
    dbus.mainloop.glib.threads_init()
    dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
except ImportError:
    pass


logger = logging.getLogger(__name__)


def main():
    log.bootstrap_delayed_logging()
    logger.info('Starting Mopidy %s', versioning.get_version())

    signal.signal(signal.SIGTERM, process.sigterm_handler)
    # Windows does not have signal.SIGUSR1
    if hasattr(signal, 'SIGUSR1'):
        signal.signal(signal.SIGUSR1, pykka.debug.log_thread_tracebacks)

    try:
        registry = ext.Registry()

        root_cmd = commands.RootCommand()
        config_cmd = commands.ConfigCommand()
        deps_cmd = commands.DepsCommand()

        root_cmd.set(extension=None, registry=registry)
        root_cmd.add_child('config', config_cmd)
        root_cmd.add_child('deps', deps_cmd)

        extensions_data = ext.load_extensions()

        for data in extensions_data:
            if data.command:  # TODO: check isinstance?
                data.command.set(extension=data.extension)
                root_cmd.add_child(data.extension.ext_name, data.command)

        args = root_cmd.parse(sys.argv[1:])

        config, config_errors = config_lib.load(
            args.config_files,
            [d.config_schema for d in extensions_data],
            [d.config_defaults for d in extensions_data],
            args.config_overrides)

        create_core_dirs(config)
        create_initial_config_file(args, extensions_data)

        log.setup_logging(
            config, args.base_verbosity_level, args.verbosity_level)

        extensions = {
            'validate': [], 'config': [], 'disabled': [], 'enabled': []}
        for data in extensions_data:
            extension = data.extension

            # TODO: factor out all of this to a helper that can be tested
            if not ext.validate_extension_data(data):
                config[extension.ext_name] = {'enabled': False}
                config_errors[extension.ext_name] = {
                    'enabled': 'extension disabled by self check.'}
                extensions['validate'].append(extension)
            elif not config[extension.ext_name]['enabled']:
                config[extension.ext_name] = {'enabled': False}
                config_errors[extension.ext_name] = {
                    'enabled': 'extension disabled by user config.'}
                extensions['disabled'].append(extension)
            elif config_errors.get(extension.ext_name):
                config[extension.ext_name]['enabled'] = False
                config_errors[extension.ext_name]['enabled'] = (
                    'extension disabled due to config errors.')
                extensions['config'].append(extension)
            else:
                extensions['enabled'].append(extension)

        log_extension_info([d.extension for d in extensions_data],
                           extensions['enabled'])

        # Config and deps commands are simply special cased for now.
        if args.command == config_cmd:
            schemas = [d.config_schema for d in extensions_data]
            return args.command.run(config, config_errors, schemas)
        elif args.command == deps_cmd:
            return args.command.run()

        check_config_errors(config, config_errors, extensions)

        if not extensions['enabled']:
            logger.error('No extension enabled, exiting...')
            sys.exit(1)

        # Read-only config from here on, please.
        proxied_config = config_lib.Proxy(config)

        if args.extension and args.extension not in extensions['enabled']:
            logger.error(
                'Unable to run command provided by disabled extension %s',
                args.extension.ext_name)
            return 1

        for extension in extensions['enabled']:
            try:
                extension.setup(registry)
            except Exception:
                # TODO: would be nice a transactional registry. But sadly this
                # is a bit tricky since our current API is giving out a mutable
                # list. We might however be able to replace this with a
                # collections.Sequence to provide a RO view.
                logger.exception('Extension %s failed during setup, this might'
                                 ' have left the registry in a bad state.',
                                 extension.ext_name)

        # Anything that wants to exit after this point must use
        # mopidy.internal.process.exit_process as actors can have been started.
        try:
            return args.command.run(args, proxied_config)
        except NotImplementedError:
            print(root_cmd.format_help())
            return 1

    except KeyboardInterrupt:
        pass
    except Exception as ex:
        logger.exception(ex)
        raise


def create_core_dirs(config):
    path.get_or_create_dir(config['core']['cache_dir'])
    path.get_or_create_dir(config['core']['config_dir'])
    path.get_or_create_dir(config['core']['data_dir'])


def create_initial_config_file(args, extensions_data):
    """Initialize whatever the last config file is with defaults"""

    config_file = args.config_files[-1]

    if os.path.exists(path.expand_path(config_file)):
        return

    try:
        default = config_lib.format_initial(extensions_data)
        path.get_or_create_file(config_file, mkdir=False, content=default)
        logger.info('Initialized %s with default config', config_file)
    except IOError as error:
        logger.warning(
            'Unable to initialize %s with default config: %s',
            config_file, encoding.locale_decode(error))


def log_extension_info(all_extensions, enabled_extensions):
    # TODO: distinguish disabled vs blocked by env?
    enabled_names = {e.ext_name for e in enabled_extensions}
    disabled_names = {e.ext_name for e in all_extensions} - enabled_names
    logger.info(
        'Enabled extensions: %s', ', '.join(enabled_names) or 'none')
    logger.info(
        'Disabled extensions: %s', ', '.join(disabled_names) or 'none')


def check_config_errors(config, errors, extensions):
    fatal_errors = []
    extension_names = {}
    all_extension_names = set()

    for state in extensions:
        extension_names[state] = {e.ext_name for e in extensions[state]}
        all_extension_names.update(extension_names[state])

    for section in sorted(errors):
        if not errors[section]:
            continue

        if section not in all_extension_names:
            logger.warning('Found fatal %s configuration errors:', section)
            fatal_errors.append(section)
        elif section in extension_names['config']:
            del errors[section]['enabled']
            logger.warning('Found %s configuration errors, the extension '
                           'has been automatically disabled:', section)
        else:
            continue

        for field, msg in errors[section].items():
            logger.warning('  %s/%s %s', section, field, msg)

    if extensions['config']:
        logger.warning('Please fix the extension configuration errors or '
                       'disable the extensions to silence these messages.')

    if fatal_errors:
        logger.error('Please fix fatal configuration errors, exiting...')
        sys.exit(1)


if __name__ == '__main__':
    sys.exit(main())



from __future__ import absolute_import, unicode_literals

import copy
import logging
import time

from mopidy import models
from mopidy.internal.models import HistoryState, HistoryTrack

logger = logging.getLogger(__name__)


class HistoryController(object):
    pykka_traversable = True

    def __init__(self):
        self._history = []

    def _add_track(self, track):
        """Add track to the playback history.

        Internal method for :class:`mopidy.core.PlaybackController`.

        :param track: track to add
        :type track: :class:`mopidy.models.Track`
        """
        if not isinstance(track, models.Track):
            raise TypeError('Only Track objects can be added to the history')

        timestamp = int(time.time() * 1000)

        name_parts = []
        if track.artists:
            name_parts.append(
                ', '.join([artist.name for artist in track.artists]))
        if track.name is not None:
            name_parts.append(track.name)
        name = ' - '.join(name_parts)
        ref = models.Ref.track(uri=track.uri, name=name)

        self._history.insert(0, (timestamp, ref))

    def get_length(self):
        """Get the number of tracks in the history.

        :returns: the history length
        :rtype: int
        """
        return len(self._history)

    def get_history(self):
        """Get the track history.

        The timestamps are milliseconds since epoch.

        :returns: the track history
        :rtype: list of (timestamp, :class:`mopidy.models.Ref`) tuples
        """
        return copy.copy(self._history)

    def _save_state(self):
        # 500 tracks a 3 minutes -> 24 hours history
        count_max = 500
        count = 1
        history_list = []
        for timestamp, track in self._history:
            history_list.append(
                HistoryTrack(timestamp=timestamp, track=track))
            count += 1
            if count_max < count:
                logger.info('Limiting history to %s tracks', count_max)
                break
        return HistoryState(history=history_list)

    def _load_state(self, state, coverage):
        if state and 'history' in coverage:
            self._history = [(h.timestamp, h.track) for h in state.history]



from __future__ import absolute_import, unicode_literals

import fnmatch
import logging
import re
import time

import pykka

from mopidy import audio as audio_lib, backend, exceptions, stream
from mopidy.audio import scan, tags
from mopidy.compat import urllib
from mopidy.internal import http, playlists
from mopidy.models import Track

logger = logging.getLogger(__name__)


class StreamBackend(pykka.ThreadingActor, backend.Backend):

    def __init__(self, config, audio):
        super(StreamBackend, self).__init__()

        self._scanner = scan.Scanner(
            timeout=config['stream']['timeout'],
            proxy_config=config['proxy'])

        self._session = http.get_requests_session(
            proxy_config=config['proxy'],
            user_agent='{}/{}'.format(
                stream.Extension.dist_name, stream.Extension.version))

        blacklist = config['stream']['metadata_blacklist']
        self._blacklist_re = re.compile(
            r'^(%s)$' % '|'.join(fnmatch.translate(u) for u in blacklist))

        self._timeout = config['stream']['timeout']

        self.library = StreamLibraryProvider(backend=self)
        self.playback = StreamPlaybackProvider(audio=audio, backend=self)
        self.playlists = None

        self.uri_schemes = audio_lib.supported_uri_schemes(
            config['stream']['protocols'])

        if 'file' in self.uri_schemes and config['file']['enabled']:
            logger.warning(
                'The stream/protocols config value includes the "file" '
                'protocol. "file" playback is now handled by Mopidy-File. '
                'Please remove it from the stream/protocols config.')
            self.uri_schemes -= {'file'}


class StreamLibraryProvider(backend.LibraryProvider):
    def lookup(self, uri):
        if urllib.parse.urlsplit(uri).scheme not in self.backend.uri_schemes:
            return []

        if self.backend._blacklist_re.match(uri):
            logger.debug('URI matched metadata lookup blacklist: %s', uri)
            return [Track(uri=uri)]

        _, scan_result = _unwrap_stream(
            uri, timeout=self.backend._timeout, scanner=self.backend._scanner,
            requests_session=self.backend._session)

        if scan_result:
            track = tags.convert_tags_to_track(scan_result.tags).replace(
                uri=uri, length=scan_result.duration)
        else:
            logger.warning('Problem looking up %s', uri)
            track = Track(uri=uri)

        return [track]


class StreamPlaybackProvider(backend.PlaybackProvider):

    def translate_uri(self, uri):
        if urllib.parse.urlsplit(uri).scheme not in self.backend.uri_schemes:
            return None

        if self.backend._blacklist_re.match(uri):
            logger.debug('URI matched metadata lookup blacklist: %s', uri)
            return uri

        unwrapped_uri, _ = _unwrap_stream(
            uri, timeout=self.backend._timeout, scanner=self.backend._scanner,
            requests_session=self.backend._session)
        return unwrapped_uri


# TODO: cleanup the return value of this.
def _unwrap_stream(uri, timeout, scanner, requests_session):
    """
    Get a stream URI from a playlist URI, ``uri``.

    Unwraps nested playlists until something that's not a playlist is found or
    the ``timeout`` is reached.
    """

    original_uri = uri
    seen_uris = set()
    deadline = time.time() + timeout

    while time.time() < deadline:
        if uri in seen_uris:
            logger.info(
                'Unwrapping stream from URI (%s) failed: '
                'playlist referenced itself', uri)
            return None, None
        else:
            seen_uris.add(uri)

        logger.debug('Unwrapping stream from URI: %s', uri)

        try:
            scan_timeout = deadline - time.time()
            if scan_timeout < 0:
                logger.info(
                    'Unwrapping stream from URI (%s) failed: '
                    'timed out in %sms', uri, timeout)
                return None, None
            scan_result = scanner.scan(uri, timeout=scan_timeout)
        except exceptions.ScannerError as exc:
            logger.debug('GStreamer failed scanning URI (%s): %s', uri, exc)
            scan_result = None

        if scan_result is not None:
            if scan_result.playable or (
                not scan_result.mime.startswith('text/') and
                not scan_result.mime.startswith('application/')
            ):
                logger.debug(
                    'Unwrapped potential %s stream: %s', scan_result.mime, uri)
                return uri, scan_result

        download_timeout = deadline - time.time()
        if download_timeout < 0:
            logger.info(
                'Unwrapping stream from URI (%s) failed: timed out in %sms',
                uri, timeout)
            return None, None
        content = http.download(
            requests_session, uri, timeout=download_timeout / 1000)

        if content is None:
            logger.info(
                'Unwrapping stream from URI (%s) failed: '
                'error downloading URI %s', original_uri, uri)
            return None, None

        uris = playlists.parse(content)
        if not uris:
            logger.debug(
                'Failed parsing URI (%s) as playlist; found potential stream.',
                uri)
            return uri, None

        # TODO Test streams and return first that seems to be playable
        logger.debug(
            'Parsed playlist (%s) and found new URI: %s', uri, uris[0])
        uri = uris[0]



# encoding: utf-8

from __future__ import absolute_import, unicode_literals

import io

from mopidy.m3u import translator
from mopidy.models import Playlist, Ref, Track


def loads(s, basedir=b'.'):
    return translator.load_items(io.StringIO(s), basedir=basedir)


def dumps(items):
    fp = io.StringIO()
    translator.dump_items(items, fp)
    return fp.getvalue()


def test_path_to_uri():
    from mopidy.m3u.translator import path_to_uri

    assert path_to_uri(b'test') == 'm3u:test'
    assert path_to_uri(b'test.m3u') == 'm3u:test.m3u'
    assert path_to_uri(b'./test.m3u') == 'm3u:test.m3u'
    assert path_to_uri(b'foo/../test.m3u') == 'm3u:test.m3u'
    assert path_to_uri(b'Test Playlist.m3u') == 'm3u:Test%20Playlist.m3u'
    assert path_to_uri(b'test.mp3', scheme='file') == 'file:///test.mp3'


def test_latin1_path_to_uri():
    path = 'æøå.m3u'.encode('latin-1')
    assert translator.path_to_uri(path) == 'm3u:%E6%F8%E5.m3u'


def test_utf8_path_to_uri():
    path = 'æøå.m3u'.encode('utf-8')
    assert translator.path_to_uri(path) == 'm3u:%C3%A6%C3%B8%C3%A5.m3u'


def test_uri_to_path():
    from mopidy.m3u.translator import uri_to_path

    assert uri_to_path('m3u:test.m3u') == b'test.m3u'
    assert uri_to_path(b'm3u:test.m3u') == b'test.m3u'
    assert uri_to_path('m3u:Test%20Playlist.m3u') == b'Test Playlist.m3u'
    assert uri_to_path(b'm3u:Test%20Playlist.m3u') == b'Test Playlist.m3u'
    assert uri_to_path('m3u:%E6%F8%E5.m3u') == b'\xe6\xf8\xe5.m3u'
    assert uri_to_path(b'm3u:%E6%F8%E5.m3u') == b'\xe6\xf8\xe5.m3u'
    assert uri_to_path('file:///test.mp3') == b'/test.mp3'
    assert uri_to_path(b'file:///test.mp3') == b'/test.mp3'


def test_name_from_path():
    from mopidy.m3u.translator import name_from_path

    assert name_from_path(b'test') == 'test'
    assert name_from_path(b'test.m3u') == 'test'
    assert name_from_path(b'../test.m3u') == 'test'


def test_path_from_name():
    from mopidy.m3u.translator import path_from_name

    assert path_from_name('test') == b'test'
    assert path_from_name('test', '.m3u') == b'test.m3u'
    assert path_from_name('foo/bar', sep='-') == b'foo-bar'


def test_path_to_ref():
    from mopidy.m3u.translator import path_to_ref

    assert path_to_ref(b'test.m3u') == Ref.playlist(
        uri='m3u:test.m3u', name='test'
    )
    assert path_to_ref(b'Test Playlist.m3u') == Ref.playlist(
        uri='m3u:Test%20Playlist.m3u', name='Test Playlist'
    )


def test_load_items():
    assert loads('') == []

    assert loads('test.mp3', basedir=b'/playlists') == [
        Ref.track(uri='file:///playlists/test.mp3', name='test')
    ]
    assert loads('../test.mp3', basedir=b'/playlists') == [
        Ref.track(uri='file:///test.mp3', name='test')
    ]
    assert loads('/test.mp3') == [
        Ref.track(uri='file:///test.mp3', name='test')
    ]
    assert loads('file:///test.mp3') == [
        Ref.track(uri='file:///test.mp3')
    ]
    assert loads('http://example.com/stream') == [
        Ref.track(uri='http://example.com/stream')
    ]

    assert loads('#EXTM3U\n#EXTINF:42,Test\nfile:///test.mp3\n') == [
        Ref.track(uri='file:///test.mp3', name='Test')
    ]
    assert loads('#EXTM3U\n#EXTINF:-1,Test\nhttp://example.com/stream\n') == [
        Ref.track(uri='http://example.com/stream', name='Test')
    ]


def test_dump_items():
    assert dumps([]) == ''
    assert dumps([Ref.track(uri='file:///test.mp3')]) == (
        'file:///test.mp3\n'
    )
    assert dumps([Ref.track(uri='file:///test.mp3', name='test')]) == (
        '#EXTM3U\n'
        '#EXTINF:-1,test\n'
        'file:///test.mp3\n'
    )
    assert dumps([Track(uri='file:///test.mp3', name='test', length=42)]) == (
        '#EXTM3U\n'
        '#EXTINF:-1,test\n'
        'file:///test.mp3\n'
    )
    assert dumps([Track(uri='http://example.com/stream')]) == (
        'http://example.com/stream\n'
    )
    assert dumps([Track(uri='http://example.com/stream', name='Test')]) == (
        '#EXTM3U\n'
        '#EXTINF:-1,Test\n'
        'http://example.com/stream\n'
    )


def test_playlist():
    from mopidy.m3u.translator import playlist

    assert playlist(b'test.m3u') == Playlist(
        uri='m3u:test.m3u',
        name='test'
    )
    assert playlist(b'test.m3u', [Ref(uri='file:///test.mp3')], 1) == Playlist(
        uri='m3u:test.m3u',
        name='test',
        tracks=[Track(uri='file:///test.mp3')],
        last_modified=1000
    )



from __future__ import absolute_import, unicode_literals

import socket
import unittest

from mock import Mock, patch, sentinel

from mopidy.internal import network


class FormatHostnameTest(unittest.TestCase):

    @patch('mopidy.internal.network.has_ipv6', True)
    def test_format_hostname_prefixes_ipv4_addresses_when_ipv6_available(self):
        network.has_ipv6 = True
        self.assertEqual(network.format_hostname('0.0.0.0'), '::ffff:0.0.0.0')
        self.assertEqual(network.format_hostname('1.0.0.1'), '::ffff:1.0.0.1')

    @patch('mopidy.internal.network.has_ipv6', False)
    def test_format_hostname_does_nothing_when_only_ipv4_available(self):
        network.has_ipv6 = False
        self.assertEqual(network.format_hostname('0.0.0.0'), '0.0.0.0')


class FormatSocketConnectionTest(unittest.TestCase):

    def test_format_socket_name(self):
        sock = Mock(spec=socket.SocketType)
        sock.family = socket.AF_INET
        sock.getsockname.return_value = (sentinel.ip, sentinel.port)
        self.assertEqual(
            network.format_socket_name(sock),
            '[{}]:{}'.format(sentinel.ip, sentinel.port))

    def test_format_socket_name_unix(self):
        sock = Mock(spec=socket.SocketType)
        sock.family = socket.AF_UNIX
        sock.getsockname.return_value = sentinel.sockname
        self.assertEqual(
            network.format_socket_name(sock),
            str(sentinel.sockname))


class TryIPv6SocketTest(unittest.TestCase):

    @patch('socket.has_ipv6', False)
    def test_system_that_claims_no_ipv6_support(self):
        self.assertFalse(network.try_ipv6_socket())

    @patch('socket.has_ipv6', True)
    @patch('socket.socket')
    def test_system_with_broken_ipv6(self, socket_mock):
        socket_mock.side_effect = IOError()
        self.assertFalse(network.try_ipv6_socket())

    @patch('socket.has_ipv6', True)
    @patch('socket.socket')
    def test_with_working_ipv6(self, socket_mock):
        socket_mock.return_value = Mock()
        self.assertTrue(network.try_ipv6_socket())


class CreateSocketTest(unittest.TestCase):

    @patch('mopidy.internal.network.has_ipv6', False)
    @patch('socket.socket')
    def test_ipv4_socket(self, socket_mock):
        network.create_tcp_socket()
        self.assertEqual(
            socket_mock.call_args[0], (socket.AF_INET, socket.SOCK_STREAM))

    @patch('mopidy.internal.network.has_ipv6', True)
    @patch('socket.socket')
    def test_ipv6_socket(self, socket_mock):
        network.create_tcp_socket()
        self.assertEqual(
            socket_mock.call_args[0], (socket.AF_INET6, socket.SOCK_STREAM))

    @unittest.SkipTest
    def test_ipv6_only_is_set(self):
        pass



from __future__ import absolute_import, unicode_literals

import errno
import os
import socket
import unittest

from mock import Mock, patch, sentinel

from mopidy import exceptions
from mopidy.internal import network
from mopidy.internal.gi import GLib

from tests import any_int


class ServerTest(unittest.TestCase):

    def setUp(self):  # noqa: N802
        self.mock = Mock(spec=network.Server)

    def test_init_calls_create_server_socket(self):
        network.Server.__init__(
            self.mock, sentinel.host, sentinel.port, sentinel.protocol)
        self.mock.create_server_socket.assert_called_once_with(
            sentinel.host, sentinel.port)
        self.mock.stop()

    def test_init_calls_register_server(self):
        sock = Mock(spec=socket.SocketType)
        sock.fileno.return_value = sentinel.fileno
        self.mock.create_server_socket.return_value = sock

        network.Server.__init__(
            self.mock, sentinel.host, sentinel.port, sentinel.protocol)
        self.mock.register_server_socket.assert_called_once_with(
            sentinel.fileno)

    def test_init_fails_on_fileno_call(self):
        sock = Mock(spec=socket.SocketType)
        sock.fileno.side_effect = socket.error
        self.mock.create_server_socket.return_value = sock

        with self.assertRaises(socket.error):
            network.Server.__init__(
                self.mock, sentinel.host, sentinel.port, sentinel.protocol)

    def test_init_stores_values_in_attributes(self):
        # This need to be a mock and no a sentinel as fileno() is called on it
        sock = Mock(spec=socket.SocketType)
        self.mock.create_server_socket.return_value = sock

        network.Server.__init__(
            self.mock, sentinel.host, sentinel.port, sentinel.protocol,
            max_connections=sentinel.max_connections, timeout=sentinel.timeout)
        self.assertEqual(sentinel.protocol, self.mock.protocol)
        self.assertEqual(sentinel.max_connections, self.mock.max_connections)
        self.assertEqual(sentinel.timeout, self.mock.timeout)
        self.assertEqual(sock, self.mock.server_socket)

    def test_create_server_socket_no_port(self):
        with self.assertRaises(exceptions.ValidationError):
            network.Server.create_server_socket(
                self.mock, str(sentinel.host), None)

    def test_create_server_socket_invalid_port(self):
        with self.assertRaises(exceptions.ValidationError):
            network.Server.create_server_socket(
                self.mock, str(sentinel.host), str(sentinel.port))

    @patch.object(network, 'create_tcp_socket', spec=socket.SocketType)
    def test_create_server_socket_sets_up_listener(self, create_tcp_socket):
        sock = create_tcp_socket.return_value

        network.Server.create_server_socket(
            self.mock, str(sentinel.host), 1234)
        sock.setblocking.assert_called_once_with(False)
        sock.bind.assert_called_once_with((str(sentinel.host), 1234))
        sock.listen.assert_called_once_with(any_int)
        create_tcp_socket.assert_called_once()

    @patch.object(network, 'create_unix_socket', spec=socket.SocketType)
    def test_create_server_socket_sets_up_listener_unix(
            self,
            create_unix_socket):
        sock = create_unix_socket.return_value

        network.Server.create_server_socket(
            self.mock, 'unix:' + str(sentinel.host), sentinel.port)
        sock.setblocking.assert_called_once_with(False)
        sock.bind.assert_called_once_with(str(sentinel.host))
        sock.listen.assert_called_once_with(any_int)
        create_unix_socket.assert_called_once()

    @patch.object(network, 'create_tcp_socket', new=Mock())
    def test_create_server_socket_fails(self):
        network.create_tcp_socket.side_effect = socket.error
        with self.assertRaises(socket.error):
            network.Server.create_server_socket(
                self.mock, str(sentinel.host), 1234)

    @patch.object(network, 'create_unix_socket', new=Mock())
    def test_create_server_socket_fails_unix(self):
        network.create_unix_socket.side_effect = socket.error
        with self.assertRaises(socket.error):
            network.Server.create_server_socket(
                self.mock, 'unix:' + str(sentinel.host), sentinel.port)

    @patch.object(network, 'create_tcp_socket', new=Mock())
    def test_create_server_bind_fails(self):
        sock = network.create_tcp_socket.return_value
        sock.bind.side_effect = socket.error

        with self.assertRaises(socket.error):
            network.Server.create_server_socket(
                self.mock, str(sentinel.host), 1234)

    @patch.object(network, 'create_unix_socket', new=Mock())
    def test_create_server_bind_fails_unix(self):
        sock = network.create_unix_socket.return_value
        sock.bind.side_effect = socket.error

        with self.assertRaises(socket.error):
            network.Server.create_server_socket(
                self.mock, 'unix:' + str(sentinel.host), sentinel.port)

    @patch.object(network, 'create_tcp_socket', new=Mock())
    def test_create_server_listen_fails(self):
        sock = network.create_tcp_socket.return_value
        sock.listen.side_effect = socket.error

        with self.assertRaises(socket.error):
            network.Server.create_server_socket(
                self.mock, str(sentinel.host), 1234)

    @patch.object(network, 'create_unix_socket', new=Mock())
    def test_create_server_listen_fails_unix(self):
        sock = network.create_unix_socket.return_value
        sock.listen.side_effect = socket.error

        with self.assertRaises(socket.error):
            network.Server.create_server_socket(
                self.mock, 'unix:' + str(sentinel.host), sentinel.port)

    @patch.object(os, 'unlink', new=Mock())
    @patch.object(GLib, 'source_remove', new=Mock())
    def test_stop_server_cleans_unix_socket(self):
        self.mock.watcher = Mock()
        sock = Mock()
        sock.family = socket.AF_UNIX
        self.mock.server_socket = sock
        network.Server.stop(self.mock)
        os.unlink.assert_called_once_with(sock.getsockname())

    @patch.object(GLib, 'io_add_watch', new=Mock())
    def test_register_server_socket_sets_up_io_watch(self):
        network.Server.register_server_socket(self.mock, sentinel.fileno)
        GLib.io_add_watch.assert_called_once_with(
            sentinel.fileno, GLib.IO_IN, self.mock.handle_connection)

    def test_handle_connection(self):
        self.mock.accept_connection.return_value = (
            sentinel.sock, sentinel.addr)
        self.mock.maximum_connections_exceeded.return_value = False

        self.assertTrue(network.Server.handle_connection(
            self.mock, sentinel.fileno, GLib.IO_IN))
        self.mock.accept_connection.assert_called_once_with()
        self.mock.maximum_connections_exceeded.assert_called_once_with()
        self.mock.init_connection.assert_called_once_with(
            sentinel.sock, sentinel.addr)
        self.assertEqual(0, self.mock.reject_connection.call_count)

    def test_handle_connection_exceeded_connections(self):
        self.mock.accept_connection.return_value = (
            sentinel.sock, sentinel.addr)
        self.mock.maximum_connections_exceeded.return_value = True

        self.assertTrue(network.Server.handle_connection(
            self.mock, sentinel.fileno, GLib.IO_IN))
        self.mock.accept_connection.assert_called_once_with()
        self.mock.maximum_connections_exceeded.assert_called_once_with()
        self.mock.reject_connection.assert_called_once_with(
            sentinel.sock, sentinel.addr)
        self.assertEqual(0, self.mock.init_connection.call_count)

    def test_accept_connection(self):
        sock = Mock(spec=socket.SocketType)
        connected_sock = Mock(spec=socket.SocketType)
        sock.accept.return_value = (connected_sock, sentinel.addr)
        self.mock.server_socket = sock

        sock, addr = network.Server.accept_connection(self.mock)
        self.assertEqual(connected_sock, sock)
        self.assertEqual(sentinel.addr, addr)

    def test_accept_connection_unix(self):
        sock = Mock(spec=socket.SocketType)
        connected_sock = Mock(spec=socket.SocketType)
        connected_sock.family = socket.AF_UNIX
        connected_sock.getsockname.return_value = sentinel.sockname
        sock.accept.return_value = (connected_sock, sentinel.addr)
        self.mock.server_socket = sock

        sock, addr = network.Server.accept_connection(self.mock)
        self.assertEqual(connected_sock, sock)
        self.assertEqual((sentinel.sockname, None), addr)

    def test_accept_connection_recoverable_error(self):
        sock = Mock(spec=socket.SocketType)
        self.mock.server_socket = sock

        for error in (errno.EAGAIN, errno.EINTR):
            sock.accept.side_effect = socket.error(error, '')
            with self.assertRaises(network.ShouldRetrySocketCall):
                network.Server.accept_connection(self.mock)

    # FIXME decide if this should be allowed to propegate
    def test_accept_connection_unrecoverable_error(self):
        sock = Mock(spec=socket.SocketType)
        self.mock.server_socket = sock
        sock.accept.side_effect = socket.error
        with self.assertRaises(socket.error):
            network.Server.accept_connection(self.mock)

    def test_maximum_connections_exceeded(self):
        self.mock.max_connections = 10

        self.mock.number_of_connections.return_value = 11
        self.assertTrue(network.Server.maximum_connections_exceeded(self.mock))

        self.mock.number_of_connections.return_value = 10
        self.assertTrue(network.Server.maximum_connections_exceeded(self.mock))

        self.mock.number_of_connections.return_value = 9
        self.assertFalse(
            network.Server.maximum_connections_exceeded(self.mock))

    @patch('pykka.ActorRegistry.get_by_class')
    def test_number_of_connections(self, get_by_class):
        self.mock.protocol = sentinel.protocol

        get_by_class.return_value = [1, 2, 3]
        self.assertEqual(3, network.Server.number_of_connections(self.mock))

        get_by_class.return_value = []
        self.assertEqual(0, network.Server.number_of_connections(self.mock))

    @patch.object(network, 'Connection', new=Mock())
    def test_init_connection(self):
        self.mock.protocol = sentinel.protocol
        self.mock.protocol_kwargs = {}
        self.mock.timeout = sentinel.timeout

        network.Server.init_connection(self.mock, sentinel.sock, sentinel.addr)
        network.Connection.assert_called_once_with(
            sentinel.protocol, {}, sentinel.sock, sentinel.addr,
            sentinel.timeout)

    @patch.object(network, 'format_socket_name', new=Mock())
    def test_reject_connection(self):
        sock = Mock(spec=socket.SocketType)

        network.Server.reject_connection(
            self.mock, sock, (sentinel.host, sentinel.port))
        sock.close.assert_called_once_with()

    @patch.object(network, 'format_socket_name', new=Mock())
    def test_reject_connection_error(self):
        sock = Mock(spec=socket.SocketType)
        sock.close.side_effect = socket.error

        network.Server.reject_connection(
            self.mock, sock, (sentinel.host, sentinel.port))
        sock.close.assert_called_once_with()



from __future__ import unicode_literals

import contextlib
import logging
import time

from mopidy.internal import log


logger = logging.getLogger(__name__)


@contextlib.contextmanager
def time_logger(name, level=log.TRACE_LOG_LEVEL):
    start = time.time()
    yield
    logger.log(level, '%s took %dms', name, (time.time() - start) * 1000)



from __future__ import absolute_import, unicode_literals

import logging
import os
import re
import stat
import string
import threading

from mopidy import compat, exceptions
from mopidy.compat import queue, urllib
from mopidy.internal import encoding, xdg


logger = logging.getLogger(__name__)


XDG_DIRS = xdg.get_dirs()


def get_or_create_dir(dir_path):
    if not isinstance(dir_path, bytes):
        raise ValueError('Path is not a bytestring.')
    dir_path = expand_path(dir_path)
    if os.path.isfile(dir_path):
        raise OSError(
            'A file with the same name as the desired dir, '
            '"%s", already exists.' % dir_path)
    elif not os.path.isdir(dir_path):
        logger.info('Creating dir %s', dir_path)
        os.makedirs(dir_path, 0o755)
    return dir_path


def get_or_create_file(file_path, mkdir=True, content=None):
    if not isinstance(file_path, bytes):
        raise ValueError('Path is not a bytestring.')
    file_path = expand_path(file_path)
    if isinstance(content, compat.text_type):
        content = content.encode('utf-8')
    if mkdir:
        get_or_create_dir(os.path.dirname(file_path))
    if not os.path.isfile(file_path):
        logger.info('Creating file %s', file_path)
        with open(file_path, 'wb') as fh:
            if content is not None:
                fh.write(content)
    return file_path


def get_unix_socket_path(socket_path):
    match = re.search('^unix:(.*)', socket_path)
    if not match:
        return None
    return match.group(1)


def path_to_uri(path):
    """
    Convert OS specific path to file:// URI.

    Accepts either unicode strings or bytestrings. The encoding of any
    bytestring will be maintained so that :func:`uri_to_path` can return the
    same bytestring.

    Returns a file:// URI as an unicode string.
    """
    if isinstance(path, compat.text_type):
        path = path.encode('utf-8')
    path = urllib.parse.quote(path)
    return urllib.parse.urlunsplit((b'file', b'', path, b'', b''))


def uri_to_path(uri):
    """
    Convert an URI to a OS specific path.

    Returns a bytestring, since the file path can contain chars with other
    encoding than UTF-8.

    If we had returned these paths as unicode strings, you wouldn't be able to
    look up the matching dir or file on your file system because the exact path
    would be lost by ignoring its encoding.
    """
    if isinstance(uri, compat.text_type):
        uri = uri.encode('utf-8')
    return urllib.parse.unquote(urllib.parse.urlsplit(uri).path)


def split_path(path):
    parts = []
    while True:
        path, part = os.path.split(path)
        if part:
            parts.insert(0, part)
        if not path or path == b'/':
            break
    return parts


def expand_path(path):
    # TODO: document as we want people to use this.
    if not isinstance(path, bytes):
        raise ValueError('Path is not a bytestring.')
    try:
        path = string.Template(path).substitute(XDG_DIRS)
    except KeyError:
        return None
    path = os.path.expanduser(path)
    path = os.path.abspath(path)
    return path


def _find_worker(relative, follow, done, work, results, errors):
    """Worker thread for collecting stat() results.

    :param str relative: directory to make results relative to
    :param bool follow: if symlinks should be followed
    :param threading.Event done: event indicating that all work has been done
    :param queue.Queue work: queue of paths to process
    :param dict results: shared dictionary for storing all the stat() results
    :param dict errors: shared dictionary for storing any per path errors
    """
    while not done.is_set():
        try:
            entry, parents = work.get(block=False)
        except queue.Empty:
            continue

        if relative:
            path = os.path.relpath(entry, relative)
        else:
            path = entry

        try:
            if follow:
                st = os.stat(entry)
            else:
                st = os.lstat(entry)

            if (st.st_dev, st.st_ino) in parents:
                errors[path] = exceptions.FindError('Sym/hardlink loop found.')
                continue

            parents = parents + [(st.st_dev, st.st_ino)]
            if stat.S_ISDIR(st.st_mode):
                for e in os.listdir(entry):
                    work.put((os.path.join(entry, e), parents))
            elif stat.S_ISREG(st.st_mode):
                results[path] = st
            elif stat.S_ISLNK(st.st_mode):
                errors[path] = exceptions.FindError('Not following symlinks.')
            else:
                errors[path] = exceptions.FindError('Not a file or directory.')

        except OSError as e:
            errors[path] = exceptions.FindError(
                encoding.locale_decode(e.strerror), e.errno)
        finally:
            work.task_done()


def _find(root, thread_count=10, relative=False, follow=False):
    """Threaded find implementation that provides stat results for files.

    Tries to protect against sym/hardlink loops by keeping an eye on parent
    (st_dev, st_ino) pairs.

    :param str root: root directory to search from, may not be a file
    :param int thread_count: number of workers to use, mainly useful to
        mitigate network lag when scanning on NFS etc.
    :param bool relative: if results should be relative to root or absolute
    :param bool follow: if symlinks should be followed
    """
    threads = []
    results = {}
    errors = {}
    done = threading.Event()
    work = queue.Queue()
    work.put((os.path.abspath(root), []))

    if not relative:
        root = None

    args = (root, follow, done, work, results, errors)
    for i in range(thread_count):
        t = threading.Thread(target=_find_worker, args=args)
        t.daemon = True
        t.start()
        threads.append(t)

    work.join()
    done.set()
    for t in threads:
        t.join()
    return results, errors


def find_mtimes(root, follow=False):
    results, errors = _find(root, relative=False, follow=follow)
    # return the mtimes as integer milliseconds
    mtimes = {f: int(st.st_mtime * 1000) for f, st in results.items()}
    return mtimes, errors


def is_path_inside_base_dir(path, base_path):
    if not isinstance(path, bytes):
        raise ValueError('path is not a bytestring')
    if not isinstance(base_path, bytes):
        raise ValueError('base_path is not a bytestring')

    if path.endswith(os.sep):
        raise ValueError('Path %s cannot end with a path separator'
                         % path)
    # Expand symlinks
    real_base_path = os.path.realpath(base_path)
    real_path = os.path.realpath(path)

    if os.path.isfile(path):
        # Use dir of file for prefix comparision, so we don't accept
        # /tmp/foo.m3u as being inside /tmp/foo, simply because they have a
        # common prefix, /tmp/foo, which matches the base path, /tmp/foo.
        real_path = os.path.dirname(real_path)

    # Check if dir of file is the base path or a subdir
    common_prefix = os.path.commonprefix([real_base_path, real_path])
    return common_prefix == real_base_path


# FIXME replace with mock usage in tests.
class Mtime(object):

    def __init__(self):
        self.fake = None

    def __call__(self, path):
        if self.fake is not None:
            return self.fake
        return int(os.stat(path).st_mtime)

    def set_fake_time(self, time):
        self.fake = time

    def undo_fake(self):
        self.fake = None


mtime = Mtime()



# XXX This module is only here to ease the migration from Mopidy-Local being
# bundled with Mopidy to being an independent extension. This file should
# probably be removed before Mopidy 3.0 final ships.

import warnings

from mopidy_local import *  # noqa


warnings.warn(
    'Mopidy-Local has been moved to its own project. '
    'Update any imports from `mopidy.local` to use `mopidy_local` instead.',
    DeprecationWarning
)



from __future__ import absolute_import, unicode_literals

import unittest

from mopidy.core import PlaybackState
from mopidy.internal import deprecation
from mopidy.models import Track

from tests.mpd import protocol


PAUSED = PlaybackState.PAUSED
PLAYING = PlaybackState.PLAYING
STOPPED = PlaybackState.STOPPED


class PlaybackOptionsHandlerTest(protocol.BaseTestCase):

    def test_consume_off(self):
        self.send_request('consume "0"')
        self.assertFalse(self.core.tracklist.get_consume().get())
        self.assertInResponse('OK')

    def test_consume_off_without_quotes(self):
        self.send_request('consume 0')
        self.assertFalse(self.core.tracklist.get_consume().get())
        self.assertInResponse('OK')

    def test_consume_on(self):
        self.send_request('consume "1"')
        self.assertTrue(self.core.tracklist.get_consume().get())
        self.assertInResponse('OK')

    def test_consume_on_without_quotes(self):
        self.send_request('consume 1')
        self.assertTrue(self.core.tracklist.get_consume().get())
        self.assertInResponse('OK')

    def test_crossfade(self):
        self.send_request('crossfade "10"')
        self.assertInResponse('ACK [0@0] {crossfade} Not implemented')

    def test_random_off(self):
        self.send_request('random "0"')
        self.assertFalse(self.core.tracklist.get_random().get())
        self.assertInResponse('OK')

    def test_random_off_without_quotes(self):
        self.send_request('random 0')
        self.assertFalse(self.core.tracklist.get_random().get())
        self.assertInResponse('OK')

    def test_random_on(self):
        self.send_request('random "1"')
        self.assertTrue(self.core.tracklist.get_random().get())
        self.assertInResponse('OK')

    def test_random_on_without_quotes(self):
        self.send_request('random 1')
        self.assertTrue(self.core.tracklist.get_random().get())
        self.assertInResponse('OK')

    def test_repeat_off(self):
        self.send_request('repeat "0"')
        self.assertFalse(self.core.tracklist.get_repeat().get())
        self.assertInResponse('OK')

    def test_repeat_off_without_quotes(self):
        self.send_request('repeat 0')
        self.assertFalse(self.core.tracklist.get_repeat().get())
        self.assertInResponse('OK')

    def test_repeat_on(self):
        self.send_request('repeat "1"')
        self.assertTrue(self.core.tracklist.get_repeat().get())
        self.assertInResponse('OK')

    def test_repeat_on_without_quotes(self):
        self.send_request('repeat 1')
        self.assertTrue(self.core.tracklist.get_repeat().get())
        self.assertInResponse('OK')

    def test_single_off(self):
        self.send_request('single "0"')
        self.assertFalse(self.core.tracklist.get_single().get())
        self.assertInResponse('OK')

    def test_single_off_without_quotes(self):
        self.send_request('single 0')
        self.assertFalse(self.core.tracklist.get_single().get())
        self.assertInResponse('OK')

    def test_single_on(self):
        self.send_request('single "1"')
        self.assertTrue(self.core.tracklist.get_single().get())
        self.assertInResponse('OK')

    def test_single_on_without_quotes(self):
        self.send_request('single 1')
        self.assertTrue(self.core.tracklist.get_single().get())
        self.assertInResponse('OK')

    def test_replay_gain_mode_off(self):
        self.send_request('replay_gain_mode "off"')
        self.assertInResponse('ACK [0@0] {replay_gain_mode} Not implemented')

    def test_replay_gain_mode_track(self):
        self.send_request('replay_gain_mode "track"')
        self.assertInResponse('ACK [0@0] {replay_gain_mode} Not implemented')

    def test_replay_gain_mode_album(self):
        self.send_request('replay_gain_mode "album"')
        self.assertInResponse('ACK [0@0] {replay_gain_mode} Not implemented')

    def test_replay_gain_status_default(self):
        self.send_request('replay_gain_status')
        self.assertInResponse('OK')
        self.assertInResponse('replay_gain_mode: off')

    def test_mixrampdb(self):
        self.send_request('mixrampdb "10"')
        self.assertInResponse('ACK [0@0] {mixrampdb} Not implemented')

    def test_mixrampdelay(self):
        self.send_request('mixrampdelay "10"')
        self.assertInResponse('ACK [0@0] {mixrampdelay} Not implemented')

    @unittest.SkipTest
    def test_replay_gain_status_off(self):
        pass

    @unittest.SkipTest
    def test_replay_gain_status_track(self):
        pass

    @unittest.SkipTest
    def test_replay_gain_status_album(self):
        pass


class PlaybackControlHandlerTest(protocol.BaseTestCase):

    def setUp(self):  # noqa: N802
        super(PlaybackControlHandlerTest, self).setUp()
        self.tracks = [Track(uri='dummy:a', length=40000),
                       Track(uri='dummy:b', length=40000)]
        self.backend.library.dummy_library = self.tracks
        self.core.tracklist.add(uris=[t.uri for t in self.tracks]).get()

    def test_next(self):
        self.core.tracklist.clear().get()
        self.send_request('next')
        self.assertInResponse('OK')

    def test_pause_off(self):
        self.send_request('play "0"')
        self.send_request('pause "1"')
        self.send_request('pause "0"')
        self.assertEqual(PLAYING, self.core.playback.get_state().get())
        self.assertInResponse('OK')

    def test_pause_on(self):
        self.send_request('play "0"')
        self.send_request('pause "1"')
        self.assertEqual(PAUSED, self.core.playback.get_state().get())
        self.assertInResponse('OK')

    def test_pause_toggle(self):
        self.send_request('play "0"')
        self.assertEqual(PLAYING, self.core.playback.get_state().get())
        self.assertInResponse('OK')

        with deprecation.ignore('mpd.protocol.playback.pause:state_arg'):
            self.send_request('pause')
            self.assertEqual(PAUSED, self.core.playback.get_state().get())
            self.assertInResponse('OK')

            self.send_request('pause')
            self.assertEqual(PLAYING, self.core.playback.get_state().get())
            self.assertInResponse('OK')

    def test_play_without_pos(self):
        self.send_request('play')
        self.assertEqual(PLAYING, self.core.playback.get_state().get())
        self.assertInResponse('OK')

    def test_play_with_pos(self):
        self.send_request('play "0"')
        self.assertEqual(PLAYING, self.core.playback.get_state().get())
        self.assertInResponse('OK')

    def test_play_with_pos_without_quotes(self):
        self.send_request('play 0')
        self.assertEqual(PLAYING, self.core.playback.get_state().get())
        self.assertInResponse('OK')

    def test_play_with_pos_out_of_bounds(self):
        self.core.tracklist.clear().get()
        self.send_request('play "0"')
        self.assertEqual(STOPPED, self.core.playback.get_state().get())
        self.assertInResponse('ACK [2@0] {play} Bad song index')

    def test_play_minus_one_plays_first_in_playlist_if_no_current_track(self):
        self.assertEqual(self.core.playback.get_current_track().get(), None)

        self.send_request('play "-1"')
        self.assertEqual(PLAYING, self.core.playback.get_state().get())
        self.assertEqual(
            'dummy:a', self.core.playback.get_current_track().get().uri)
        self.assertInResponse('OK')

    def test_play_minus_one_plays_current_track_if_current_track_is_set(self):
        self.assertEqual(self.core.playback.get_current_track().get(), None)
        self.core.playback.play()
        self.core.playback.next()
        self.core.playback.stop().get()
        self.assertNotEqual(self.core.playback.get_current_track().get(), None)

        self.send_request('play "-1"')
        self.assertEqual(PLAYING, self.core.playback.get_state().get())
        self.assertEqual(
            'dummy:b', self.core.playback.get_current_track().get().uri)
        self.assertInResponse('OK')

    def test_play_minus_one_on_empty_playlist_does_not_ack(self):
        self.core.tracklist.clear()

        self.send_request('play "-1"')
        self.assertEqual(STOPPED, self.core.playback.get_state().get())
        self.assertEqual(None, self.core.playback.get_current_track().get())
        self.assertInResponse('OK')

    def test_play_minus_is_ignored_if_playing(self):
        self.core.playback.play().get()
        self.core.playback.seek(30000)
        self.assertGreaterEqual(
            self.core.playback.get_time_position().get(), 30000)
        self.assertEqual(PLAYING, self.core.playback.get_state().get())

        self.send_request('play "-1"')
        self.assertEqual(PLAYING, self.core.playback.get_state().get())
        self.assertGreaterEqual(
            self.core.playback.get_time_position().get(), 30000)
        self.assertInResponse('OK')

    def test_play_minus_one_resumes_if_paused(self):
        self.core.playback.play().get()
        self.core.playback.seek(30000)
        self.assertGreaterEqual(
            self.core.playback.get_time_position().get(), 30000)
        self.assertEqual(PLAYING, self.core.playback.get_state().get())
        self.core.playback.pause()
        self.assertEqual(PAUSED, self.core.playback.get_state().get())

        self.send_request('play "-1"')
        self.assertEqual(PLAYING, self.core.playback.get_state().get())
        self.assertGreaterEqual(
            self.core.playback.get_time_position().get(), 30000)
        self.assertInResponse('OK')

    def test_playid(self):
        self.send_request('playid "1"')
        self.assertEqual(PLAYING, self.core.playback.get_state().get())
        self.assertInResponse('OK')

    def test_playid_without_quotes(self):
        self.send_request('playid 1')
        self.assertEqual(PLAYING, self.core.playback.get_state().get())
        self.assertInResponse('OK')

    def test_playid_minus_1_plays_first_in_playlist_if_no_current_track(self):
        self.assertEqual(self.core.playback.get_current_track().get(), None)

        self.send_request('playid "-1"')
        self.assertEqual(PLAYING, self.core.playback.get_state().get())
        self.assertEqual(
            'dummy:a', self.core.playback.get_current_track().get().uri)
        self.assertInResponse('OK')

    def test_playid_minus_1_plays_current_track_if_current_track_is_set(self):
        self.assertEqual(self.core.playback.get_current_track().get(), None)
        self.core.playback.play().get()
        self.core.playback.next().get()
        self.core.playback.stop()
        self.assertNotEqual(None, self.core.playback.get_current_track().get())

        self.send_request('playid "-1"')
        self.assertEqual(PLAYING, self.core.playback.get_state().get())
        self.assertEqual(
            'dummy:b', self.core.playback.get_current_track().get().uri)
        self.assertInResponse('OK')

    def test_playid_minus_one_on_empty_playlist_does_not_ack(self):
        self.core.tracklist.clear()

        self.send_request('playid "-1"')
        self.assertEqual(STOPPED, self.core.playback.get_state().get())
        self.assertEqual(None, self.core.playback.get_current_track().get())
        self.assertInResponse('OK')

    def test_playid_minus_is_ignored_if_playing(self):
        self.core.playback.play().get()
        self.core.playback.seek(30000)
        self.assertGreaterEqual(
            self.core.playback.get_time_position().get(), 30000)
        self.assertEqual(PLAYING, self.core.playback.get_state().get())

        self.send_request('playid "-1"')
        self.assertEqual(PLAYING, self.core.playback.get_state().get())
        self.assertGreaterEqual(
            self.core.playback.get_time_position().get(), 30000)
        self.assertInResponse('OK')

    def test_playid_minus_one_resumes_if_paused(self):
        self.core.playback.play().get()
        self.core.playback.seek(30000)
        self.assertGreaterEqual(
            self.core.playback.get_time_position().get(), 30000)
        self.assertEqual(PLAYING, self.core.playback.get_state().get())
        self.core.playback.pause()
        self.assertEqual(PAUSED, self.core.playback.get_state().get())

        self.send_request('playid "-1"')
        self.assertEqual(PLAYING, self.core.playback.get_state().get())
        self.assertGreaterEqual(
            self.core.playback.get_time_position().get(), 30000)
        self.assertInResponse('OK')

    def test_playid_which_does_not_exist(self):
        self.send_request('playid "12345"')
        self.assertInResponse('ACK [50@0] {playid} No such song')

    def test_previous(self):
        self.core.tracklist.clear().get()
        self.send_request('previous')
        self.assertInResponse('OK')

    def test_seek_in_current_track(self):
        self.core.playback.play()

        self.send_request('seek "0" "30"')

        current_track = self.core.playback.get_current_track().get()
        self.assertEqual(current_track, self.tracks[0])
        self.assertGreaterEqual(self.core.playback.get_time_position(), 30000)
        self.assertInResponse('OK')

    def test_seek_in_another_track(self):
        self.core.playback.play()
        current_track = self.core.playback.get_current_track().get()
        self.assertNotEqual(current_track, self.tracks[1])

        self.send_request('seek "1" "30"')

        current_track = self.core.playback.get_current_track().get()
        self.assertEqual(current_track, self.tracks[1])
        self.assertInResponse('OK')

    def test_seek_without_quotes(self):
        self.core.playback.play()

        self.send_request('seek 0 30')
        self.assertGreaterEqual(
            self.core.playback.get_time_position().get(), 30000)
        self.assertInResponse('OK')

    def test_seekid_in_current_track(self):
        self.core.playback.play()

        self.send_request('seekid "1" "30"')

        current_track = self.core.playback.get_current_track().get()
        self.assertEqual(current_track, self.tracks[0])
        self.assertGreaterEqual(
            self.core.playback.get_time_position().get(), 30000)
        self.assertInResponse('OK')

    def test_seekid_in_another_track(self):
        self.core.playback.play()

        self.send_request('seekid "2" "30"')

        current_tl_track = self.core.playback.get_current_tl_track().get()
        self.assertEqual(current_tl_track.tlid, 2)
        self.assertEqual(current_tl_track.track, self.tracks[1])
        self.assertInResponse('OK')

    def test_seekcur_absolute_value(self):
        self.core.playback.play().get()

        self.send_request('seekcur "30"')

        self.assertGreaterEqual(
            self.core.playback.get_time_position().get(), 30000)
        self.assertInResponse('OK')

    def test_seekcur_positive_diff(self):
        self.core.playback.play().get()
        self.core.playback.seek(10000)
        self.assertGreaterEqual(
            self.core.playback.get_time_position().get(), 10000)

        self.send_request('seekcur "+20"')

        self.assertGreaterEqual(
            self.core.playback.get_time_position().get(), 30000)
        self.assertInResponse('OK')

    def test_seekcur_negative_diff(self):
        self.core.playback.play().get()
        self.core.playback.seek(30000)
        self.assertGreaterEqual(
            self.core.playback.get_time_position().get(), 30000)

        self.send_request('seekcur "-20"')

        self.assertLessEqual(
            self.core.playback.get_time_position().get(), 15000)
        self.assertInResponse('OK')

    def test_stop(self):
        self.core.tracklist.clear().get()
        self.send_request('stop')
        self.assertEqual(STOPPED, self.core.playback.get_state().get())
        self.assertInResponse('OK')


class VolumeTest(protocol.BaseTestCase):

    def test_setvol_below_min(self):
        self.send_request('setvol "-10"')
        self.assertEqual(0, self.core.mixer.get_volume().get())
        self.assertInResponse('OK')

    def test_setvol_min(self):
        self.send_request('setvol "0"')
        self.assertEqual(0, self.core.mixer.get_volume().get())
        self.assertInResponse('OK')

    def test_setvol_middle(self):
        self.send_request('setvol "50"')
        self.assertEqual(50, self.core.mixer.get_volume().get())
        self.assertInResponse('OK')

    def test_setvol_max(self):
        self.send_request('setvol "100"')
        self.assertEqual(100, self.core.mixer.get_volume().get())
        self.assertInResponse('OK')

    def test_setvol_above_max(self):
        self.send_request('setvol "110"')
        self.assertEqual(100, self.core.mixer.get_volume().get())
        self.assertInResponse('OK')

    def test_setvol_plus_is_ignored(self):
        self.send_request('setvol "+10"')
        self.assertEqual(10, self.core.mixer.get_volume().get())
        self.assertInResponse('OK')

    def test_setvol_without_quotes(self):
        self.send_request('setvol 50')
        self.assertEqual(50, self.core.mixer.get_volume().get())
        self.assertInResponse('OK')

    def test_volume_plus(self):
        self.core.mixer.set_volume(50)

        self.send_request('volume +20')

        self.assertEqual(70, self.core.mixer.get_volume().get())
        self.assertInResponse('OK')

    def test_volume_minus(self):
        self.core.mixer.set_volume(50)

        self.send_request('volume -20')

        self.assertEqual(30, self.core.mixer.get_volume().get())
        self.assertInResponse('OK')

    def test_volume_less_than_minus_100(self):
        self.core.mixer.set_volume(50)

        self.send_request('volume -110')

        self.assertEqual(50, self.core.mixer.get_volume().get())
        self.assertInResponse('ACK [2@0] {volume} Invalid volume value')

    def test_volume_more_than_plus_100(self):
        self.core.mixer.set_volume(50)

        self.send_request('volume +110')

        self.assertEqual(50, self.core.mixer.get_volume().get())
        self.assertInResponse('ACK [2@0] {volume} Invalid volume value')


class VolumeWithNoMixerTest(protocol.BaseTestCase):
    enable_mixer = False

    def test_setvol_without_mixer_fails(self):
        self.send_request('setvol "100"')
        self.assertInResponse('ACK [52@0] {setvol} problems setting volume')

    def test_volume_without_mixer_failes(self):
        self.send_request('volume +100')
        self.assertInResponse('ACK [52@0] {volume} problems setting volume')



from __future__ import absolute_import, unicode_literals

import logging
import logging.config
import logging.handlers
import platform


LOG_LEVELS = {
    -1: dict(root=logging.ERROR, mopidy=logging.WARNING),
    0: dict(root=logging.ERROR, mopidy=logging.INFO),
    1: dict(root=logging.WARNING, mopidy=logging.DEBUG),
    2: dict(root=logging.INFO, mopidy=logging.DEBUG),
    3: dict(root=logging.DEBUG, mopidy=logging.DEBUG),
    4: dict(root=logging.NOTSET, mopidy=logging.NOTSET),
}

# Custom log level which has even lower priority than DEBUG
TRACE_LOG_LEVEL = 5
logging.addLevelName(TRACE_LOG_LEVEL, 'TRACE')

logger = logging.getLogger(__name__)


class DelayedHandler(logging.Handler):

    def __init__(self):
        logging.Handler.__init__(self)
        self._released = False
        self._buffer = []

    def handle(self, record):
        if not self._released:
            self._buffer.append(record)

    def release(self):
        self._released = True
        root = logging.getLogger('')
        while self._buffer:
            root.handle(self._buffer.pop(0))


_delayed_handler = DelayedHandler()


def bootstrap_delayed_logging():
    root = logging.getLogger('')
    root.setLevel(logging.NOTSET)
    root.addHandler(_delayed_handler)


def setup_logging(config, base_verbosity_level, args_verbosity_level):
    logging.captureWarnings(True)

    if config['logging']['config_file']:
        # Logging config from file must be read before other handlers are
        # added. If not, the other handlers will have no effect.
        try:
            path = config['logging']['config_file']
            logging.config.fileConfig(path, disable_existing_loggers=False)
        except Exception as e:
            # Catch everything as logging does not specify what can go wrong.
            logger.error('Loading logging config %r failed. %s', path, e)

    loglevels = config.get('loglevels', {})

    verbosity_level = get_verbosity_level(
        config, base_verbosity_level, args_verbosity_level)
    verbosity_filter = VerbosityFilter(verbosity_level, loglevels)

    formatter = logging.Formatter(config['logging']['format'])

    if config['logging']['color']:
        handler = ColorizingStreamHandler(config.get('logcolors', {}))
    else:
        handler = logging.StreamHandler()
    handler.addFilter(verbosity_filter)
    handler.setFormatter(formatter)

    logging.getLogger('').addHandler(handler)

    _delayed_handler.release()


def get_verbosity_level(config, base_verbosity_level, args_verbosity_level):
    if args_verbosity_level:
        result = base_verbosity_level + args_verbosity_level
    else:
        result = base_verbosity_level + config['logging']['verbosity']

    if result < min(LOG_LEVELS.keys()):
        result = min(LOG_LEVELS.keys())
    if result > max(LOG_LEVELS.keys()):
        result = max(LOG_LEVELS.keys())

    return result


class VerbosityFilter(logging.Filter):

    def __init__(self, verbosity_level, loglevels):
        self.verbosity_level = verbosity_level
        self.loglevels = loglevels

    def filter(self, record):
        for name, required_log_level in self.loglevels.items():
            if record.name == name or record.name.startswith(name + '.'):
                return record.levelno >= required_log_level

        if record.name.startswith('mopidy'):
            required_log_level = LOG_LEVELS[self.verbosity_level]['mopidy']
        else:
            required_log_level = LOG_LEVELS[self.verbosity_level]['root']
        return record.levelno >= required_log_level


#: Available log colors.
COLORS = [b'black', b'red', b'green', b'yellow', b'blue', b'magenta', b'cyan',
          b'white']


class ColorizingStreamHandler(logging.StreamHandler):

    """
    Stream handler which colorizes the log using ANSI escape sequences.

    Does nothing on Windows, which doesn't support ANSI escape sequences.

    This implementation is based upon https://gist.github.com/vsajip/758430,
    which is:

        Copyright (C) 2010-2012 Vinay Sajip. All rights reserved.
        Licensed under the new BSD license.
    """

    # Map logging levels to (background, foreground, bold/intense)
    level_map = {
        TRACE_LOG_LEVEL: (None, 'blue', False),
        logging.DEBUG: (None, 'blue', False),
        logging.INFO: (None, 'white', False),
        logging.WARNING: (None, 'yellow', False),
        logging.ERROR: (None, 'red', False),
        logging.CRITICAL: ('red', 'white', True),
    }
    # Map logger name to foreground colors
    logger_map = {}

    csi = '\x1b['
    reset = '\x1b[0m'

    is_windows = platform.system() == 'Windows'

    def __init__(self, logger_colors):
        super(ColorizingStreamHandler, self).__init__()
        self.logger_map = logger_colors

    @property
    def is_tty(self):
        isatty = getattr(self.stream, 'isatty', None)
        return isatty and isatty()

    def emit(self, record):
        try:
            message = self.format(record)
            self.stream.write(message)
            self.stream.write(getattr(self, 'terminator', '\n'))
            self.flush()
        except Exception:
            self.handleError(record)

    def format(self, record):
        message = logging.StreamHandler.format(self, record)
        if not self.is_tty or self.is_windows:
            return message
        for name, color in self.logger_map.iteritems():
            if record.name.startswith(name):
                return self.colorize(message, fg=color)
        if record.levelno in self.level_map:
            bg, fg, bold = self.level_map[record.levelno]
            return self.colorize(message, bg=bg, fg=fg, bold=bold)
        return message

    def colorize(self, message, bg=None, fg=None, bold=False):
        params = []
        if bg in COLORS:
            params.append(str(COLORS.index(bg) + 40))
        if fg in COLORS:
            params.append(str(COLORS.index(fg) + 30))
        if bold:
            params.append('1')
        if params:
            message = ''.join((
                self.csi, ';'.join(params), 'm', message, self.reset))
        return message



from __future__ import absolute_import, unicode_literals

import mock

import pytest

from mopidy.mpd import actor

# NOTE: Should be kept in sync with all events from mopidy.core.listener


@pytest.mark.parametrize("event,expected", [
    (['track_playback_paused', 'tl_track', 'time_position'], None),
    (['track_playback_resumed', 'tl_track', 'time_position'], None),
    (['track_playback_started', 'tl_track'], None),
    (['track_playback_ended', 'tl_track', 'time_position'], None),
    (['playback_state_changed', 'old_state', 'new_state'], 'player'),
    (['tracklist_changed'], 'playlist'),
    (['playlists_loaded'], 'stored_playlist'),
    (['playlist_changed', 'playlist'], 'stored_playlist'),
    (['playlist_deleted', 'uri'], 'stored_playlist'),
    (['options_changed'], 'options'),
    (['volume_changed', 'volume'], 'mixer'),
    (['mute_changed', 'mute'], 'output'),
    (['seeked', 'time_position'], 'player'),
    (['stream_title_changed', 'title'], 'playlist'),
])
def test_idle_hooked_up_correctly(event, expected):
    config = {'mpd': {'hostname': 'foobar',
                      'port': 1234,
                      'zeroconf': None,
                      'max_connections': None,
                      'connection_timeout': None}}

    with mock.patch.object(actor.MpdFrontend, '_setup_server'):
        frontend = actor.MpdFrontend(core=mock.Mock(), config=config)

    with mock.patch('mopidy.listener.send') as send_mock:
        frontend.on_event(event[0], **{e: None for e in event[1:]})

    if expected is None:
        assert not send_mock.call_args
    else:
        send_mock.assert_called_once_with(mock.ANY, expected)



from __future__ import absolute_import, unicode_literals

from mopidy.models import Track

from tests.mpd import protocol


class StatusHandlerTest(protocol.BaseTestCase):

    def test_clearerror(self):
        self.send_request('clearerror')
        self.assertEqualResponse('ACK [0@0] {clearerror} Not implemented')

    def test_currentsong(self):
        track = Track(uri='dummy:/a')
        self.backend.library.dummy_library = [track]
        self.core.tracklist.add(uris=[track.uri]).get()

        self.core.playback.play().get()
        self.send_request('currentsong')
        self.assertInResponse('file: dummy:/a')
        self.assertInResponse('Time: 0')
        self.assertNotInResponse('Artist: ')
        self.assertNotInResponse('Title: ')
        self.assertNotInResponse('Album: ')
        self.assertNotInResponse('Track: 0')
        self.assertNotInResponse('Date: ')
        self.assertInResponse('Pos: 0')
        self.assertInResponse('Id: 1')
        self.assertInResponse('OK')

    def test_currentsong_without_song(self):
        self.send_request('currentsong')
        self.assertInResponse('OK')

    def test_stats_command(self):
        self.send_request('stats')
        self.assertInResponse('OK')

    def test_status_command(self):
        self.send_request('status')
        self.assertInResponse('OK')



from __future__ import absolute_import, unicode_literals

import contextlib
import logging

from mopidy import exceptions
from mopidy.compat import urllib
from mopidy.core import listener
from mopidy.internal import validation
from mopidy.models import Playlist, Ref

logger = logging.getLogger(__name__)


@contextlib.contextmanager
def _backend_error_handling(backend, reraise=None):
    try:
        yield
    except exceptions.ValidationError as e:
        logger.error('%s backend returned bad data: %s',
                     backend.actor_ref.actor_class.__name__, e)
    except Exception as e:
        if reraise and isinstance(e, reraise):
            raise
        logger.exception('%s backend caused an exception.',
                         backend.actor_ref.actor_class.__name__)


class PlaylistsController(object):
    pykka_traversable = True

    def __init__(self, backends, core):
        self.backends = backends
        self.core = core

    def get_uri_schemes(self):
        """
        Get the list of URI schemes that support playlists.

        :rtype: list of string

        .. versionadded:: 2.0
        """
        return list(sorted(self.backends.with_playlists.keys()))

    def as_list(self):
        """
        Get a list of the currently available playlists.

        Returns a list of :class:`~mopidy.models.Ref` objects referring to the
        playlists. In other words, no information about the playlists' content
        is given.

        :rtype: list of :class:`mopidy.models.Ref`

        .. versionadded:: 1.0
        """
        futures = {
            backend: backend.playlists.as_list()
            for backend in set(self.backends.with_playlists.values())}

        results = []
        for b, future in futures.items():
            try:
                with _backend_error_handling(b, reraise=NotImplementedError):
                    playlists = future.get()
                    if playlists is not None:
                        validation.check_instances(playlists, Ref)
                        results.extend(playlists)
            except NotImplementedError:
                backend_name = b.actor_ref.actor_class.__name__
                logger.warning(
                    '%s does not implement playlists.as_list(). '
                    'Please upgrade it.', backend_name)

        return results

    def get_items(self, uri):
        """
        Get the items in a playlist specified by ``uri``.

        Returns a list of :class:`~mopidy.models.Ref` objects referring to the
        playlist's items.

        If a playlist with the given ``uri`` doesn't exist, it returns
        :class:`None`.

        :rtype: list of :class:`mopidy.models.Ref`, or :class:`None`

        .. versionadded:: 1.0
        """
        validation.check_uri(uri)

        uri_scheme = urllib.parse.urlparse(uri).scheme
        backend = self.backends.with_playlists.get(uri_scheme, None)

        if not backend:
            return None

        with _backend_error_handling(backend):
            items = backend.playlists.get_items(uri).get()
            items is None or validation.check_instances(items, Ref)
            return items

        return None

    def create(self, name, uri_scheme=None):
        """
        Create a new playlist.

        If ``uri_scheme`` matches an URI scheme handled by a current backend,
        that backend is asked to create the playlist. If ``uri_scheme`` is
        :class:`None` or doesn't match a current backend, the first backend is
        asked to create the playlist.

        All new playlists must be created by calling this method, and **not**
        by creating new instances of :class:`mopidy.models.Playlist`.

        :param name: name of the new playlist
        :type name: string
        :param uri_scheme: use the backend matching the URI scheme
        :type uri_scheme: string
        :rtype: :class:`mopidy.models.Playlist` or :class:`None`
        """
        if uri_scheme in self.backends.with_playlists:
            backends = [self.backends.with_playlists[uri_scheme]]
        else:
            backends = self.backends.with_playlists.values()

        for backend in backends:
            with _backend_error_handling(backend):
                result = backend.playlists.create(name).get()
                if result is None:
                    continue
                validation.check_instance(result, Playlist)
                listener.CoreListener.send('playlist_changed', playlist=result)
                return result

        return None

    def delete(self, uri):
        """
        Delete playlist identified by the URI.

        If the URI doesn't match the URI schemes handled by the current
        backends, nothing happens.

        Returns :class:`True` if deleted, :class:`False` otherwise.

        :param uri: URI of the playlist to delete
        :type uri: string
        :rtype: :class:`bool`

        .. versionchanged:: 2.2
            Return type defined.
        """
        validation.check_uri(uri)

        uri_scheme = urllib.parse.urlparse(uri).scheme
        backend = self.backends.with_playlists.get(uri_scheme, None)
        if not backend:
            return False

        success = False
        with _backend_error_handling(backend):
            success = backend.playlists.delete(uri).get()

        if success is None:
            # Return type was defined in Mopidy 2.2. Assume everything went
            # well if the backend doesn't report otherwise.
            success = True

        if success:
            listener.CoreListener.send('playlist_deleted', uri=uri)

        return success

    def lookup(self, uri):
        """
        Lookup playlist with given URI in both the set of playlists and in any
        other playlist sources. Returns :class:`None` if not found.

        :param uri: playlist URI
        :type uri: string
        :rtype: :class:`mopidy.models.Playlist` or :class:`None`
        """
        uri_scheme = urllib.parse.urlparse(uri).scheme
        backend = self.backends.with_playlists.get(uri_scheme, None)
        if not backend:
            return None

        with _backend_error_handling(backend):
            playlist = backend.playlists.lookup(uri).get()
            playlist is None or validation.check_instance(playlist, Playlist)
            return playlist

        return None

    # TODO: there is an inconsistency between library.refresh(uri) and this
    # call, not sure how to sort this out.
    def refresh(self, uri_scheme=None):
        """
        Refresh the playlists in :attr:`playlists`.

        If ``uri_scheme`` is :class:`None`, all backends are asked to refresh.
        If ``uri_scheme`` is an URI scheme handled by a backend, only that
        backend is asked to refresh. If ``uri_scheme`` doesn't match any
        current backend, nothing happens.

        :param uri_scheme: limit to the backend matching the URI scheme
        :type uri_scheme: string
        """
        # TODO: check: uri_scheme is None or uri_scheme?

        futures = {}
        backends = {}
        playlists_loaded = False

        for backend_scheme, backend in self.backends.with_playlists.items():
            backends.setdefault(backend, set()).add(backend_scheme)

        for backend, backend_schemes in backends.items():
            if uri_scheme is None or uri_scheme in backend_schemes:
                futures[backend] = backend.playlists.refresh()

        for backend, future in futures.items():
            with _backend_error_handling(backend):
                future.get()
                playlists_loaded = True

        if playlists_loaded:
            listener.CoreListener.send('playlists_loaded')

    def save(self, playlist):
        """
        Save the playlist.

        For a playlist to be saveable, it must have the ``uri`` attribute set.
        You must not set the ``uri`` atribute yourself, but use playlist
        objects returned by :meth:`create` or retrieved from :attr:`playlists`,
        which will always give you saveable playlists.

        The method returns the saved playlist. The return playlist may differ
        from the saved playlist. E.g. if the playlist name was changed, the
        returned playlist may have a different URI. The caller of this method
        must throw away the playlist sent to this method, and use the
        returned playlist instead.

        If the playlist's URI isn't set or doesn't match the URI scheme of a
        current backend, nothing is done and :class:`None` is returned.

        :param playlist: the playlist
        :type playlist: :class:`mopidy.models.Playlist`
        :rtype: :class:`mopidy.models.Playlist` or :class:`None`
        """
        validation.check_instance(playlist, Playlist)

        if playlist.uri is None:
            return  # TODO: log this problem?

        uri_scheme = urllib.parse.urlparse(playlist.uri).scheme
        backend = self.backends.with_playlists.get(uri_scheme, None)
        if not backend:
            return None

        # TODO: we let AssertionError error through due to legacy tests :/
        with _backend_error_handling(backend, reraise=AssertionError):
            playlist = backend.playlists.save(playlist).get()
            playlist is None or validation.check_instance(playlist, Playlist)
            if playlist:
                listener.CoreListener.send(
                    'playlist_changed', playlist=playlist)
            return playlist

        return None



# encoding: utf-8

from __future__ import absolute_import, unicode_literals

import os
import platform
import shutil
import tempfile
import unittest

import pykka

from mopidy import core
from mopidy.m3u.backend import M3UBackend
from mopidy.models import Playlist, Track

from tests import dummy_audio, path_to_data_dir
from tests.m3u import generate_song


class M3UPlaylistsProviderTest(unittest.TestCase):
    backend_class = M3UBackend
    config = {
        'm3u': {
            'enabled': True,
            'base_dir': None,
            'default_encoding': 'latin-1',
            'default_extension': '.m3u',
            'playlists_dir': path_to_data_dir(''),
        }
    }

    def setUp(self):  # noqa: N802
        self.config['m3u']['playlists_dir'] = tempfile.mkdtemp()
        self.playlists_dir = self.config['m3u']['playlists_dir']
        self.base_dir = self.config['m3u']['base_dir'] or self.playlists_dir

        audio = dummy_audio.create_proxy()
        backend = M3UBackend.start(
            config=self.config, audio=audio).proxy()
        self.core = core.Core(backends=[backend])

    def tearDown(self):  # noqa: N802
        pykka.ActorRegistry.stop_all()

        if os.path.exists(self.playlists_dir):
            shutil.rmtree(self.playlists_dir)

    def test_created_playlist_is_persisted(self):
        uri = 'm3u:test.m3u'
        path = os.path.join(self.playlists_dir, b'test.m3u')
        self.assertFalse(os.path.exists(path))

        playlist = self.core.playlists.create('test')
        self.assertEqual('test', playlist.name)
        self.assertEqual(uri, playlist.uri)
        self.assertTrue(os.path.exists(path))

    def test_create_sanitizes_playlist_name(self):
        playlist = self.core.playlists.create('  ../../test FOO baR ')
        self.assertEqual('..|..|test FOO baR', playlist.name)
        path = os.path.join(self.playlists_dir, b'..|..|test FOO baR.m3u')
        self.assertEqual(self.playlists_dir, os.path.dirname(path))
        self.assertTrue(os.path.exists(path))

    def test_saved_playlist_is_persisted(self):
        uri1 = 'm3u:test1.m3u'
        uri2 = 'm3u:test2.m3u'

        path1 = os.path.join(self.playlists_dir, b'test1.m3u')
        path2 = os.path.join(self.playlists_dir, b'test2.m3u')

        playlist = self.core.playlists.create('test1')
        self.assertEqual('test1', playlist.name)
        self.assertEqual(uri1, playlist.uri)
        self.assertTrue(os.path.exists(path1))
        self.assertFalse(os.path.exists(path2))

        playlist = self.core.playlists.save(playlist.replace(name='test2'))
        self.assertEqual('test2', playlist.name)
        self.assertEqual(uri2, playlist.uri)
        self.assertFalse(os.path.exists(path1))
        self.assertTrue(os.path.exists(path2))

    def test_deleted_playlist_is_removed(self):
        uri = 'm3u:test.m3u'
        path = os.path.join(self.playlists_dir, b'test.m3u')

        self.assertFalse(os.path.exists(path))

        playlist = self.core.playlists.create('test')
        self.assertEqual('test', playlist.name)
        self.assertEqual(uri, playlist.uri)
        self.assertTrue(os.path.exists(path))

        success = self.core.playlists.delete(playlist.uri)
        self.assertTrue(success)
        self.assertFalse(os.path.exists(path))

    def test_delete_on_path_outside_playlist_dir_returns_none(self):
        success = self.core.playlists.delete('m3u:///etc/passwd')

        self.assertFalse(success)

    def test_playlist_contents_is_written_to_disk(self):
        track = Track(uri=generate_song(1))
        playlist = self.core.playlists.create('test')
        playlist = self.core.playlists.save(playlist.replace(tracks=[track]))
        path = os.path.join(self.playlists_dir, b'test.m3u')

        with open(path) as f:
            contents = f.read()

        self.assertEqual(track.uri, contents.strip())

    def test_extended_playlist_contents_is_written_to_disk(self):
        track = Track(uri=generate_song(1), name='Test', length=60000)
        playlist = self.core.playlists.create('test')
        playlist = self.core.playlists.save(playlist.replace(tracks=[track]))
        path = os.path.join(self.playlists_dir, b'test.m3u')

        with open(path) as f:
            m3u = f.read().splitlines()

        self.assertEqual(['#EXTM3U', '#EXTINF:-1,Test', track.uri], m3u)

    def test_latin1_playlist_contents_is_written_to_disk(self):
        track = Track(uri=generate_song(1), name='Test\x9f', length=60000)
        playlist = self.core.playlists.create('test')
        playlist = self.core.playlists.save(playlist.replace(tracks=[track]))
        path = os.path.join(self.playlists_dir, b'test.m3u')

        with open(path, 'rb') as f:
            m3u = f.read().splitlines()
        self.assertEqual([b'#EXTM3U', b'#EXTINF:-1,Test\x9f', track.uri], m3u)

    def test_utf8_playlist_contents_is_replaced_and_written_to_disk(self):
        track = Track(uri=generate_song(1), name='Test\u07b4', length=60000)
        playlist = self.core.playlists.create('test')
        playlist = self.core.playlists.save(playlist.replace(tracks=[track]))
        path = os.path.join(self.playlists_dir, b'test.m3u')

        with open(path, 'rb') as f:
            m3u = f.read().splitlines()
        self.assertEqual([b'#EXTM3U', b'#EXTINF:-1,Test?', track.uri], m3u)

    def test_playlists_are_loaded_at_startup(self):
        track = Track(uri='dummy:track:path2')
        playlist = self.core.playlists.create('test')
        playlist = playlist.replace(tracks=[track])
        playlist = self.core.playlists.save(playlist)

        self.assertEqual(len(self.core.playlists.as_list()), 1)
        result = self.core.playlists.lookup(playlist.uri)
        self.assertEqual(playlist.uri, result.uri)
        self.assertEqual(playlist.name, result.name)
        self.assertEqual(track.uri, result.tracks[0].uri)

    @unittest.skipIf(
        platform.system() == 'Darwin',
        'macOS 10.13 raises IOError "Illegal byte sequence" on open.')
    def test_load_playlist_with_nonfilesystem_encoding_of_filename(self):
        path = os.path.join(self.playlists_dir, 'øæå.m3u'.encode('latin-1'))
        with open(path, 'wb+') as f:
            f.write(b'#EXTM3U\n')

        self.core.playlists.refresh()

        self.assertEqual(len(self.core.playlists.as_list()), 1)
        result = self.core.playlists.as_list()
        self.assertEqual('\ufffd\ufffd\ufffd', result[0].name)

    @unittest.SkipTest
    def test_playlists_dir_is_created(self):
        pass

    def test_create_returns_playlist_with_name_set(self):
        playlist = self.core.playlists.create('test')
        self.assertEqual(playlist.name, 'test')

    def test_create_returns_playlist_with_uri_set(self):
        playlist = self.core.playlists.create('test')
        self.assert_(playlist.uri)

    def test_create_adds_playlist_to_playlists_collection(self):
        playlist = self.core.playlists.create('test')
        playlists = self.core.playlists.as_list()
        self.assertIn(playlist.uri, [ref.uri for ref in playlists])

    def test_as_list_empty_to_start_with(self):
        self.assertEqual(len(self.core.playlists.as_list()), 0)

    def test_delete_non_existant_playlist(self):
        self.core.playlists.delete('m3u:unknown')

    def test_delete_playlist_removes_it_from_the_collection(self):
        playlist = self.core.playlists.create('test')
        self.assertEqual(playlist, self.core.playlists.lookup(playlist.uri))

        self.core.playlists.delete(playlist.uri)

        self.assertIsNone(self.core.playlists.lookup(playlist.uri))

    def test_delete_playlist_without_file(self):
        playlist = self.core.playlists.create('test')
        self.assertEqual(playlist, self.core.playlists.lookup(playlist.uri))

        path = os.path.join(self.playlists_dir, b'test.m3u')
        self.assertTrue(os.path.exists(path))

        os.remove(path)
        self.assertFalse(os.path.exists(path))

        self.core.playlists.delete(playlist.uri)
        self.assertIsNone(self.core.playlists.lookup(playlist.uri))

    def test_lookup_finds_playlist_by_uri(self):
        original_playlist = self.core.playlists.create('test')

        looked_up_playlist = self.core.playlists.lookup(original_playlist.uri)

        self.assertEqual(original_playlist, looked_up_playlist)

    def test_lookup_on_path_outside_playlist_dir_returns_none(self):
        result = self.core.playlists.lookup('m3u:///etc/passwd')

        self.assertIsNone(result)

    def test_refresh(self):
        playlist = self.core.playlists.create('test')
        self.assertEqual(playlist, self.core.playlists.lookup(playlist.uri))

        self.core.playlists.refresh()

        self.assertEqual(playlist, self.core.playlists.lookup(playlist.uri))

    def test_save_replaces_existing_playlist_with_updated_playlist(self):
        playlist1 = self.core.playlists.create('test1')
        self.assertEqual(playlist1, self.core.playlists.lookup(playlist1.uri))

        playlist2 = playlist1.replace(name='test2')
        playlist2 = self.core.playlists.save(playlist2)
        self.assertIsNone(self.core.playlists.lookup(playlist1.uri))
        self.assertEqual(playlist2, self.core.playlists.lookup(playlist2.uri))

    def test_create_replaces_existing_playlist_with_updated_playlist(self):
        track = Track(uri=generate_song(1))
        playlist1 = self.core.playlists.create('test')
        playlist1 = self.core.playlists.save(playlist1.replace(tracks=[track]))
        self.assertEqual(playlist1, self.core.playlists.lookup(playlist1.uri))

        playlist2 = self.core.playlists.create('test')
        self.assertEqual(playlist1.uri, playlist2.uri)
        self.assertNotEqual(
            playlist1, self.core.playlists.lookup(playlist1.uri))
        self.assertEqual(playlist2, self.core.playlists.lookup(playlist1.uri))

    def test_save_playlist_with_new_uri(self):
        uri = 'm3u:test.m3u'
        self.core.playlists.save(Playlist(uri=uri))
        path = os.path.join(self.playlists_dir, b'test.m3u')
        self.assertTrue(os.path.exists(path))

    def test_save_on_path_outside_playlist_dir_returns_none(self):
        result = self.core.playlists.save(Playlist(uri='m3u:///tmp/test.m3u'))
        self.assertIsNone(result)

    def test_playlist_with_unknown_track(self):
        track = Track(uri='file:///dev/null')
        playlist = self.core.playlists.create('test')
        playlist = playlist.replace(tracks=[track])
        playlist = self.core.playlists.save(playlist)

        self.assertEqual(len(self.core.playlists.as_list()), 1)
        result = self.core.playlists.lookup('m3u:test.m3u')
        self.assertEqual('m3u:test.m3u', result.uri)
        self.assertEqual(playlist.name, result.name)
        self.assertEqual(track.uri, result.tracks[0].uri)

    def test_playlist_with_absolute_path(self):
        track = Track(uri='/tmp/test.mp3')
        filepath = b'/tmp/test.mp3'
        playlist = self.core.playlists.create('test')
        playlist = playlist.replace(tracks=[track])
        playlist = self.core.playlists.save(playlist)

        self.assertEqual(len(self.core.playlists.as_list()), 1)
        result = self.core.playlists.lookup('m3u:test.m3u')
        self.assertEqual('m3u:test.m3u', result.uri)
        self.assertEqual(playlist.name, result.name)
        self.assertEqual('file://' + filepath, result.tracks[0].uri)

    def test_playlist_with_relative_path(self):
        track = Track(uri='test.mp3')
        filepath = os.path.join(self.base_dir, b'test.mp3')
        playlist = self.core.playlists.create('test')
        playlist = playlist.replace(tracks=[track])
        playlist = self.core.playlists.save(playlist)

        self.assertEqual(len(self.core.playlists.as_list()), 1)
        result = self.core.playlists.lookup('m3u:test.m3u')
        self.assertEqual('m3u:test.m3u', result.uri)
        self.assertEqual(playlist.name, result.name)
        self.assertEqual('file://' + filepath, result.tracks[0].uri)

    def test_playlist_sort_order(self):
        def check_order(playlists, names):
            self.assertEqual(names, [playlist.name for playlist in playlists])

        self.core.playlists.create('c')
        self.core.playlists.create('a')
        self.core.playlists.create('b')

        check_order(self.core.playlists.as_list(), ['a', 'b', 'c'])

        self.core.playlists.refresh()

        check_order(self.core.playlists.as_list(), ['a', 'b', 'c'])

        playlist = self.core.playlists.lookup('m3u:a.m3u')
        playlist = playlist.replace(name='d')
        playlist = self.core.playlists.save(playlist)

        check_order(self.core.playlists.as_list(), ['b', 'c', 'd'])

        self.core.playlists.delete('m3u:c.m3u')

        check_order(self.core.playlists.as_list(), ['b', 'd'])

    def test_get_items_returns_item_refs(self):
        track = Track(uri='dummy:a', name='A', length=60000)
        playlist = self.core.playlists.create('test')
        playlist = self.core.playlists.save(playlist.replace(tracks=[track]))

        item_refs = self.core.playlists.get_items(playlist.uri)

        self.assertEqual(len(item_refs), 1)
        self.assertEqual(item_refs[0].type, 'track')
        self.assertEqual(item_refs[0].uri, 'dummy:a')
        self.assertEqual(item_refs[0].name, 'A')

    def test_get_items_of_unknown_playlist_returns_none(self):
        item_refs = self.core.playlists.get_items('dummy:unknown')

        self.assertIsNone(item_refs)

    def test_get_items_from_file_outside_playlist_dir_returns_none(self):
        item_refs = self.core.playlists.get_items('m3u:///etc/passwd')

        self.assertIsNone(item_refs)


class M3UPlaylistsProviderBaseDirectoryTest(M3UPlaylistsProviderTest):

    def setUp(self):  # noqa: N802
        self.config['m3u']['base_dir'] = tempfile.mkdtemp()
        super(M3UPlaylistsProviderBaseDirectoryTest, self).setUp()



from __future__ import absolute_import, unicode_literals

import collections

from mopidy import compat, exceptions
from mopidy.compat import urllib

PLAYBACK_STATES = {'paused', 'stopped', 'playing'}

SEARCH_FIELDS = {
    'uri', 'track_name', 'album', 'artist', 'albumartist', 'composer',
    'performer', 'track_no', 'genre', 'date', 'comment', 'any'}

PLAYLIST_FIELDS = {'uri', 'name'}  # TODO: add length and last_modified?

TRACKLIST_FIELDS = {  # TODO: add bitrate, length, disc_no, track_no, modified?
    'uri', 'name', 'genre', 'date', 'comment', 'musicbrainz_id'}

DISTINCT_FIELDS = {
    'track', 'artist', 'albumartist', 'album', 'composer', 'performer', 'date',
    'genre'}


# TODO: _check_iterable(check, msg, **kwargs) + [check(a) for a in arg]?
def _check_iterable(arg, msg, **kwargs):
    """Ensure we have an iterable which is not a string or an iterator"""
    if isinstance(arg, compat.string_types):
        raise exceptions.ValidationError(msg.format(arg=arg, **kwargs))
    elif not isinstance(arg, collections.Iterable):
        raise exceptions.ValidationError(msg.format(arg=arg, **kwargs))
    elif iter(arg) is iter(arg):
        raise exceptions.ValidationError(msg.format(arg=arg, **kwargs))


def check_choice(arg, choices, msg='Expected one of {choices}, not {arg!r}'):
    if arg not in choices:
        raise exceptions.ValidationError(msg.format(
            arg=arg, choices=tuple(choices)))


def check_boolean(arg, msg='Expected a boolean, not {arg!r}'):
    check_instance(arg, bool, msg=msg)


def check_instance(arg, cls, msg='Expected a {name} instance, not {arg!r}'):
    if not isinstance(arg, cls):
        raise exceptions.ValidationError(
            msg.format(arg=arg, name=cls.__name__))


def check_instances(arg, cls, msg='Expected a list of {name}, not {arg!r}'):
    _check_iterable(arg, msg, name=cls.__name__)
    if not all(isinstance(instance, cls) for instance in arg):
        raise exceptions.ValidationError(
            msg.format(arg=arg, name=cls.__name__))


def check_integer(arg, min=None, max=None):
    if not isinstance(arg, compat.integer_types):
        raise exceptions.ValidationError('Expected an integer, not %r' % arg)
    elif min is not None and arg < min:
        raise exceptions.ValidationError(
            'Expected number larger or equal to %d, not %r' % (min, arg))
    elif max is not None and arg > max:
        raise exceptions.ValidationError(
            'Expected number smaller or equal to %d, not %r' % (max, arg))


def check_query(arg, fields=SEARCH_FIELDS, list_values=True):
    # TODO: normalize name  -> track_name
    # TODO: normalize value -> [value]
    # TODO: normalize blank -> [] or just remove field?
    # TODO: remove list_values?

    if not isinstance(arg, collections.Mapping):
        raise exceptions.ValidationError(
            'Expected a query dictionary, not {arg!r}'.format(arg=arg))

    for key, value in arg.items():
        check_choice(key, fields, msg='Expected query field to be one of '
                     '{choices}, not {arg!r}')
        if list_values:
            msg = 'Expected "{key}" to be list of strings, not {arg!r}'
            _check_iterable(value, msg, key=key)
            [_check_query_value(key, v, msg) for v in value]
        else:
            _check_query_value(
                key, value, 'Expected "{key}" to be a string, not {arg!r}')


def _check_query_value(key, arg, msg):
    if not isinstance(arg, compat.string_types) or not arg.strip():
        raise exceptions.ValidationError(msg.format(arg=arg, key=key))


def check_uri(arg, msg='Expected a valid URI, not {arg!r}'):
    if not isinstance(arg, compat.string_types):
        raise exceptions.ValidationError(msg.format(arg=arg))
    elif urllib.parse.urlparse(arg).scheme == '':
        raise exceptions.ValidationError(msg.format(arg=arg))


def check_uris(arg, msg='Expected a list of URIs, not {arg!r}'):
    _check_iterable(arg, msg)
    [check_uri(a, msg) for a in arg]



from __future__ import absolute_import, unicode_literals

import re

# TOOD: refactor this into a generic mapper that does not know about browse
# or playlists and then use one instance for each case?


class MpdUriMapper(object):

    """
    Maintains the mappings between uniquified MPD names and URIs.
    """

    #: The Mopidy core API. An instance of :class:`mopidy.core.Core`.
    core = None

    _invalid_browse_chars = re.compile(r'[\n\r]')
    _invalid_playlist_chars = re.compile(r'[/]')

    def __init__(self, core=None):
        self.core = core
        self._uri_from_name = {}
        self._browse_name_from_uri = {}
        self._playlist_name_from_uri = {}

    def _create_unique_name(self, name, uri):
        stripped_name = self._invalid_browse_chars.sub(' ', name)
        name = stripped_name
        i = 2
        while name in self._uri_from_name:
            if self._uri_from_name[name] == uri:
                return name
            name = '%s [%d]' % (stripped_name, i)
            i += 1
        return name

    def insert(self, name, uri, playlist=False):
        """
        Create a unique and MPD compatible name that maps to the given URI.
        """
        name = self._create_unique_name(name, uri)
        self._uri_from_name[name] = uri
        if playlist:
            self._playlist_name_from_uri[uri] = name
        else:
            self._browse_name_from_uri[uri] = name
        return name

    def uri_from_name(self, name):
        """
        Return the uri for the given MPD name.
        """
        return self._uri_from_name.get(name)

    def refresh_playlists_mapping(self):
        """
        Maintain map between playlists and unique playlist names to be used by
        MPD.
        """
        if self.core is None:
            return

        for playlist_ref in self.core.playlists.as_list().get():
            if not playlist_ref.name:
                continue
            name = self._invalid_playlist_chars.sub('|', playlist_ref.name)
            self.insert(name, playlist_ref.uri, playlist=True)

    def playlist_uri_from_name(self, name):
        """
        Helper function to retrieve a playlist URI from its unique MPD name.
        """
        if name not in self._uri_from_name:
            self.refresh_playlists_mapping()
        return self._uri_from_name.get(name)

    def playlist_name_from_uri(self, uri):
        """
        Helper function to retrieve the unique MPD playlist name from its URI.
        """
        if uri not in self._playlist_name_from_uri:
            self.refresh_playlists_mapping()
        return self._playlist_name_from_uri[uri]



from __future__ import absolute_import, unicode_literals

import logging

import pykka

from mopidy import backend
from mopidy.file import library


logger = logging.getLogger(__name__)


class FileBackend(pykka.ThreadingActor, backend.Backend):
    uri_schemes = ['file']

    def __init__(self, config, audio):
        super(FileBackend, self).__init__()
        self.library = library.FileLibraryProvider(backend=self, config=config)
        self.playback = backend.PlaybackProvider(audio=audio, backend=self)
        self.playlists = None



from __future__ import absolute_import, unicode_literals

from mock import patch

from mopidy.mpd.protocol.status import SUBSYSTEMS

from tests.mpd import protocol


class IdleHandlerTest(protocol.BaseTestCase):

    def idle_event(self, subsystem):
        self.session.on_event(subsystem)

    def assertEqualEvents(self, events):  # noqa: N802
        self.assertEqual(set(events), self.context.events)

    def assertEqualSubscriptions(self, events):  # noqa: N802
        self.assertEqual(set(events), self.context.subscriptions)

    def assertNoEvents(self):  # noqa: N802
        self.assertEqualEvents([])

    def assertNoSubscriptions(self):  # noqa: N802
        self.assertEqualSubscriptions([])

    def test_base_state(self):
        self.assertNoSubscriptions()
        self.assertNoEvents()
        self.assertNoResponse()

    def test_idle(self):
        self.send_request('idle')
        self.assertEqualSubscriptions(SUBSYSTEMS)
        self.assertNoEvents()
        self.assertNoResponse()

    def test_idle_disables_timeout(self):
        self.send_request('idle')
        self.connection.disable_timeout.assert_called_once_with()

    def test_noidle(self):
        self.send_request('noidle')
        self.assertNoSubscriptions()
        self.assertNoEvents()
        self.assertNoResponse()

    def test_idle_player(self):
        self.send_request('idle player')
        self.assertEqualSubscriptions(['player'])
        self.assertNoEvents()
        self.assertNoResponse()

    def test_idle_output(self):
        self.send_request('idle output')
        self.assertEqualSubscriptions(['output'])
        self.assertNoEvents()
        self.assertNoResponse()

    def test_idle_player_playlist(self):
        self.send_request('idle player playlist')
        self.assertEqualSubscriptions(['player', 'playlist'])
        self.assertNoEvents()
        self.assertNoResponse()

    def test_idle_then_noidle(self):
        self.send_request('idle')
        self.send_request('noidle')
        self.assertNoSubscriptions()
        self.assertNoEvents()
        self.assertOnceInResponse('OK')

    def test_idle_then_noidle_enables_timeout(self):
        self.send_request('idle')
        self.send_request('noidle')
        self.connection.enable_timeout.assert_called_once_with()

    def test_idle_then_play(self):
        with patch.object(self.session, 'stop') as stop_mock:
            self.send_request('idle')
            self.send_request('play')
            stop_mock.assert_called_once_with()

    def test_idle_then_idle(self):
        with patch.object(self.session, 'stop') as stop_mock:
            self.send_request('idle')
            self.send_request('idle')
            stop_mock.assert_called_once_with()

    def test_idle_player_then_play(self):
        with patch.object(self.session, 'stop') as stop_mock:
            self.send_request('idle player')
            self.send_request('play')
            stop_mock.assert_called_once_with()

    def test_idle_then_player(self):
        self.send_request('idle')
        self.idle_event('player')
        self.assertNoSubscriptions()
        self.assertNoEvents()
        self.assertOnceInResponse('changed: player')
        self.assertOnceInResponse('OK')

    def test_idle_player_then_event_player(self):
        self.send_request('idle player')
        self.idle_event('player')
        self.assertNoSubscriptions()
        self.assertNoEvents()
        self.assertOnceInResponse('changed: player')
        self.assertOnceInResponse('OK')

    def test_idle_then_output(self):
        self.send_request('idle')
        self.idle_event('output')
        self.assertNoSubscriptions()
        self.assertNoEvents()
        self.assertOnceInResponse('changed: output')
        self.assertOnceInResponse('OK')

    def test_idle_output_then_event_output(self):
        self.send_request('idle output')
        self.idle_event('output')
        self.assertNoSubscriptions()
        self.assertNoEvents()
        self.assertOnceInResponse('changed: output')
        self.assertOnceInResponse('OK')

    def test_idle_player_then_noidle(self):
        self.send_request('idle player')
        self.send_request('noidle')
        self.assertNoSubscriptions()
        self.assertNoEvents()
        self.assertOnceInResponse('OK')

    def test_idle_player_playlist_then_noidle(self):
        self.send_request('idle player playlist')
        self.send_request('noidle')
        self.assertNoEvents()
        self.assertNoSubscriptions()
        self.assertOnceInResponse('OK')

    def test_idle_player_playlist_then_player(self):
        self.send_request('idle player playlist')
        self.idle_event('player')
        self.assertNoEvents()
        self.assertNoSubscriptions()
        self.assertOnceInResponse('changed: player')
        self.assertNotInResponse('changed: playlist')
        self.assertOnceInResponse('OK')

    def test_idle_playlist_then_player(self):
        self.send_request('idle playlist')
        self.idle_event('player')
        self.assertEqualEvents(['player'])
        self.assertEqualSubscriptions(['playlist'])
        self.assertNoResponse()

    def test_idle_playlist_then_player_then_playlist(self):
        self.send_request('idle playlist')
        self.idle_event('player')
        self.idle_event('playlist')
        self.assertNoEvents()
        self.assertNoSubscriptions()
        self.assertNotInResponse('changed: player')
        self.assertOnceInResponse('changed: playlist')
        self.assertOnceInResponse('OK')

    def test_player(self):
        self.idle_event('player')
        self.assertEqualEvents(['player'])
        self.assertNoSubscriptions()
        self.assertNoResponse()

    def test_player_then_idle_player(self):
        self.idle_event('player')
        self.send_request('idle player')
        self.assertNoEvents()
        self.assertNoSubscriptions()
        self.assertOnceInResponse('changed: player')
        self.assertNotInResponse('changed: playlist')
        self.assertOnceInResponse('OK')

    def test_player_then_playlist(self):
        self.idle_event('player')
        self.idle_event('playlist')
        self.assertEqualEvents(['player', 'playlist'])
        self.assertNoSubscriptions()
        self.assertNoResponse()

    def test_player_then_idle(self):
        self.idle_event('player')
        self.send_request('idle')
        self.assertNoEvents()
        self.assertNoSubscriptions()
        self.assertOnceInResponse('changed: player')
        self.assertOnceInResponse('OK')

    def test_player_then_playlist_then_idle(self):
        self.idle_event('player')
        self.idle_event('playlist')
        self.send_request('idle')
        self.assertNoEvents()
        self.assertNoSubscriptions()
        self.assertOnceInResponse('changed: player')
        self.assertOnceInResponse('changed: playlist')
        self.assertOnceInResponse('OK')

    def test_player_then_idle_playlist(self):
        self.idle_event('player')
        self.send_request('idle playlist')
        self.assertEqualEvents(['player'])
        self.assertEqualSubscriptions(['playlist'])
        self.assertNoResponse()

    def test_player_then_idle_playlist_then_noidle(self):
        self.idle_event('player')
        self.send_request('idle playlist')
        self.send_request('noidle')
        self.assertNoEvents()
        self.assertNoSubscriptions()
        self.assertOnceInResponse('OK')

    def test_player_then_playlist_then_idle_playlist(self):
        self.idle_event('player')
        self.idle_event('playlist')
        self.send_request('idle playlist')
        self.assertNoEvents()
        self.assertNoSubscriptions()
        self.assertNotInResponse('changed: player')
        self.assertOnceInResponse('changed: playlist')
        self.assertOnceInResponse('OK')

    def test_output_then_idle_toggleoutput(self):
        self.idle_event('output')
        self.send_request('idle output')
        self.assertNoEvents()
        self.assertNoSubscriptions()
        self.assertOnceInResponse('changed: output')
        self.assertOnceInResponse('OK')



from __future__ import absolute_import, unicode_literals

import os
import unittest

import mock

import tornado.testing
import tornado.web
import tornado.websocket

import mopidy
from mopidy.http import handlers


class StaticFileHandlerTest(tornado.testing.AsyncHTTPTestCase):

    def get_app(self):
        return tornado.web.Application([
            (r'/(.*)', handlers.StaticFileHandler, {
                'path': os.path.dirname(__file__),
                'default_filename': 'test_handlers.py'
            })
        ])

    def test_static_handler(self):
        response = self.fetch('/test_handlers.py', method='GET')

        self.assertEqual(200, response.code)
        self.assertEqual(
            response.headers['X-Mopidy-Version'], mopidy.__version__)
        self.assertEqual(
            response.headers['Cache-Control'], 'no-cache')

    def test_static_default_filename(self):
        response = self.fetch('/', method='GET')

        self.assertEqual(200, response.code)
        self.assertEqual(
            response.headers['X-Mopidy-Version'], mopidy.__version__)
        self.assertEqual(
            response.headers['Cache-Control'], 'no-cache')


class WebSocketHandlerTest(tornado.testing.AsyncHTTPTestCase):

    def get_app(self):
        self.core = mock.Mock()
        return tornado.web.Application([
            (r'/ws/?', handlers.WebSocketHandler, {
                'core': self.core, 'allowed_origins': [],
                'csrf_protection': True
            })
        ])

    def connection(self):
        url = self.get_url('/ws').replace('http', 'ws')
        return tornado.websocket.websocket_connect(url, self.io_loop)

    @tornado.testing.gen_test
    def test_invalid_json_rpc_request_doesnt_crash_handler(self):
        # An uncaught error would result in no message, so this is just a
        # simplistic test to verify this.
        conn = yield self.connection()
        conn.write_message('invalid request')
        message = yield conn.read_message()
        self.assertTrue(message)

    @tornado.testing.gen_test
    def test_broadcast_makes_it_to_client(self):
        conn = yield self.connection()
        handlers.WebSocketHandler.broadcast('message')
        message = yield conn.read_message()
        self.assertEqual(message, 'message')

    @tornado.testing.gen_test
    def test_broadcast_to_client_that_just_closed_connection(self):
        conn = yield self.connection()
        conn.stream.close()
        handlers.WebSocketHandler.broadcast('message')

    @tornado.testing.gen_test
    def test_broadcast_to_client_without_ws_connection_present(self):
        yield self.connection()
        # Tornado checks for ws_connection and raises WebSocketClosedError
        # if it is missing, this test case simulates winning a race were
        # this has happened but we have not yet been removed from clients.
        for client in handlers.WebSocketHandler.clients:
            client.ws_connection = None
        handlers.WebSocketHandler.broadcast('message')


class CheckOriginTests(unittest.TestCase):

    def setUp(self):
        self.headers = {'Host': 'localhost:6680'}
        self.allowed = set()

    def test_missing_origin_blocked(self):
        self.assertFalse(handlers.check_origin(
            None, self.headers, self.allowed))

    def test_empty_origin_allowed(self):
        self.assertTrue(handlers.check_origin('', self.headers, self.allowed))

    def test_chrome_file_origin_allowed(self):
        self.assertTrue(handlers.check_origin(
            'file://', self.headers, self.allowed))

    def test_firefox_null_origin_allowed(self):
        self.assertTrue(handlers.check_origin(
            'null', self.headers, self.allowed))

    def test_same_host_origin_allowed(self):
        self.assertTrue(handlers.check_origin(
            'http://localhost:6680', self.headers, self.allowed))

    def test_different_host_origin_blocked(self):
        self.assertFalse(handlers.check_origin(
            'http://other:6680', self.headers, self.allowed))

    def test_different_port_blocked(self):
        self.assertFalse(handlers.check_origin(
            'http://localhost:80', self.headers, self.allowed))

    def test_extra_origin_allowed(self):
        self.allowed.add('other:6680')
        self.assertTrue(handlers.check_origin(
            'http://other:6680', self.headers, self.allowed))



from __future__ import absolute_import, unicode_literals

import unittest

import mock

import pykka

from mopidy import core, mixer
from mopidy.internal.models import MixerState

from tests import dummy_mixer


class CoreMixerTest(unittest.TestCase):

    def setUp(self):  # noqa: N802
        self.mixer = mock.Mock(spec=mixer.Mixer)
        self.core = core.Core(mixer=self.mixer, backends=[])

    def test_get_volume(self):
        self.mixer.get_volume.return_value.get.return_value = 30

        self.assertEqual(self.core.mixer.get_volume(), 30)
        self.mixer.get_volume.assert_called_once_with()

    def test_set_volume(self):
        self.mixer.set_volume.return_value.get.return_value = True
        self.core.mixer.set_volume(30)

        self.mixer.set_volume.assert_called_once_with(30)

    def test_get_mute(self):
        self.mixer.get_mute.return_value.get.return_value = True

        self.assertEqual(self.core.mixer.get_mute(), True)
        self.mixer.get_mute.assert_called_once_with()

    def test_set_mute(self):
        self.mixer.set_mute.return_value.get.return_value = True
        self.core.mixer.set_mute(True)

        self.mixer.set_mute.assert_called_once_with(True)


class CoreNoneMixerTest(unittest.TestCase):

    def setUp(self):  # noqa: N802
        self.core = core.Core(mixer=None, backends=[])

    def test_get_volume_return_none_because_it_is_unknown(self):
        self.assertEqual(self.core.mixer.get_volume(), None)

    def test_set_volume_return_false_because_it_failed(self):
        self.assertEqual(self.core.mixer.set_volume(30), False)

    def test_get_mute_return_none_because_it_is_unknown(self):
        self.assertEqual(self.core.mixer.get_mute(), None)

    def test_set_mute_return_false_because_it_failed(self):
        self.assertEqual(self.core.mixer.set_mute(True), False)


@mock.patch.object(mixer.MixerListener, 'send')
class CoreMixerListenerTest(unittest.TestCase):

    def setUp(self):  # noqa: N802
        self.mixer = dummy_mixer.create_proxy()
        self.core = core.Core(mixer=self.mixer, backends=[])

    def tearDown(self):  # noqa: N802
        pykka.ActorRegistry.stop_all()

    def test_forwards_mixer_volume_changed_event_to_frontends(self, send):
        self.assertEqual(self.core.mixer.set_volume(volume=60), True)
        self.assertEqual(send.call_args[0][0], 'volume_changed')
        self.assertEqual(send.call_args[1]['volume'], 60)

    def test_forwards_mixer_mute_changed_event_to_frontends(self, send):
        self.core.mixer.set_mute(mute=True)

        self.assertEqual(send.call_args[0][0], 'mute_changed')
        self.assertEqual(send.call_args[1]['mute'], True)


@mock.patch.object(mixer.MixerListener, 'send')
class CoreNoneMixerListenerTest(unittest.TestCase):

    def setUp(self):  # noqa: N802
        self.core = core.Core(mixer=None, backends=[])

    def test_forwards_mixer_volume_changed_event_to_frontends(self, send):
        self.assertEqual(self.core.mixer.set_volume(volume=60), False)
        self.assertEqual(send.call_count, 0)

    def test_forwards_mixer_mute_changed_event_to_frontends(self, send):
        self.core.mixer.set_mute(mute=True)
        self.assertEqual(send.call_count, 0)


class MockBackendCoreMixerBase(unittest.TestCase):

    def setUp(self):  # noqa: N802
        self.mixer = mock.Mock()
        self.mixer.actor_ref.actor_class.__name__ = 'DummyMixer'
        self.core = core.Core(mixer=self.mixer, backends=[])


class GetVolumeBadBackendTest(MockBackendCoreMixerBase):

    def test_backend_raises_exception(self):
        self.mixer.get_volume.return_value.get.side_effect = Exception
        self.assertEqual(self.core.mixer.get_volume(), None)

    def test_backend_returns_too_small_value(self):
        self.mixer.get_volume.return_value.get.return_value = -1
        self.assertEqual(self.core.mixer.get_volume(), None)

    def test_backend_returns_too_large_value(self):
        self.mixer.get_volume.return_value.get.return_value = 1000
        self.assertEqual(self.core.mixer.get_volume(), None)

    def test_backend_returns_wrong_type(self):
        self.mixer.get_volume.return_value.get.return_value = '12'
        self.assertEqual(self.core.mixer.get_volume(), None)


class SetVolumeBadBackendTest(MockBackendCoreMixerBase):

    def test_backend_raises_exception(self):
        self.mixer.set_volume.return_value.get.side_effect = Exception
        self.assertFalse(self.core.mixer.set_volume(30))

    def test_backend_returns_wrong_type(self):
        self.mixer.set_volume.return_value.get.return_value = 'done'
        self.assertFalse(self.core.mixer.set_volume(30))


class GetMuteBadBackendTest(MockBackendCoreMixerBase):

    def test_backend_raises_exception(self):
        self.mixer.get_mute.return_value.get.side_effect = Exception
        self.assertEqual(self.core.mixer.get_mute(), None)

    def test_backend_returns_wrong_type(self):
        self.mixer.get_mute.return_value.get.return_value = '12'
        self.assertEqual(self.core.mixer.get_mute(), None)


class SetMuteBadBackendTest(MockBackendCoreMixerBase):

    def test_backend_raises_exception(self):
        self.mixer.set_mute.return_value.get.side_effect = Exception
        self.assertFalse(self.core.mixer.set_mute(True))

    def test_backend_returns_wrong_type(self):
        self.mixer.set_mute.return_value.get.return_value = 'done'
        self.assertFalse(self.core.mixer.set_mute(True))


class CoreMixerSaveLoadStateTest(unittest.TestCase):

    def setUp(self):  # noqa: N802
        self.mixer = dummy_mixer.create_proxy()
        self.core = core.Core(mixer=self.mixer, backends=[])

    def test_save_mute(self):
        volume = 32
        mute = False
        target = MixerState(volume=volume, mute=mute)
        self.core.mixer.set_volume(volume)
        self.core.mixer.set_mute(mute)
        value = self.core.mixer._save_state()
        self.assertEqual(target, value)

    def test_save_unmute(self):
        volume = 33
        mute = True
        target = MixerState(volume=volume, mute=mute)
        self.core.mixer.set_volume(volume)
        self.core.mixer.set_mute(mute)
        value = self.core.mixer._save_state()
        self.assertEqual(target, value)

    def test_load(self):
        self.core.mixer.set_volume(11)
        volume = 45
        target = MixerState(volume=volume)
        coverage = ['mixer']
        self.core.mixer._load_state(target, coverage)
        self.assertEqual(volume, self.core.mixer.get_volume())

    def test_load_not_covered(self):
        self.core.mixer.set_volume(21)
        self.core.mixer.set_mute(True)
        target = MixerState(volume=56, mute=False)
        coverage = ['other']
        self.core.mixer._load_state(target, coverage)
        self.assertEqual(21, self.core.mixer.get_volume())
        self.assertEqual(True, self.core.mixer.get_mute())

    def test_load_mute_on(self):
        self.core.mixer.set_mute(False)
        self.assertEqual(False, self.core.mixer.get_mute())
        target = MixerState(mute=True)
        coverage = ['mixer']
        self.core.mixer._load_state(target, coverage)
        self.assertEqual(True, self.core.mixer.get_mute())

    def test_load_mute_off(self):
        self.core.mixer.set_mute(True)
        self.assertEqual(True, self.core.mixer.get_mute())
        target = MixerState(mute=False)
        coverage = ['mixer']
        self.core.mixer._load_state(target, coverage)
        self.assertEqual(False, self.core.mixer.get_mute())

    def test_load_invalid_type(self):
        with self.assertRaises(TypeError):
            self.core.mixer._load_state(11, None)

    def test_load_none(self):
        self.core.mixer._load_state(None, None)
